/*
 client.cc
 Katsuki Ohto
 */

// 参考 http://ameblo.jp/systemtelescope/entry-11820880734.html

#include "../../../lib/picojson.h"

#include "./aiwolf.hpp"

#include "./thousand.hpp"

#include "./json.hpp"

//AIWolf::AI ai;
AIWolf::ThousandAI ai;

int sendMessage(int sock, const std::string& msg){
    char sbuffer[256];
    memset(sbuffer, 0, sizeof(sbuffer));
    snprintf(sbuffer, sizeof(sbuffer), msg.c_str());
    if(strlen(sbuffer) > 0){
        //cerr << "Client << " << sbuffer;
    }
    int err = send(sock, sbuffer, strlen(sbuffer) + 1, 0);
    if(err < 0){
        CERR << "failed to send message(" << err << ")." << endl;
    }
    return err;
}

int main(int argc, char* argv[]){
    
    // 基本的な初期化
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    
    std::string host = "127.0.0.1";
    unsigned int port = 10000;
    std::string param_dir = "./";
    
    for(int c = 1; c < argc; ++c){
        if(strstr(argv[c], "-h")){
            host = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-i")){
            param_dir = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-p")){
            port = atoi(argv[c + 1]);
        }
    }
    
    // AI設定
    ai.setName(MY_NAME);
    //ai.setHopeRole(AIWolf::Role::SEER);
    //ai.setHopeRole(AIWolf::Role::MEDIUM);
    //ai.setHopeRole(AIWolf::Role::BODYGUARD);
    //ai.setHopeRole(AIWolf::Role::WEREWOLF);
    //ai.setHopeRole(AIWolf::Role::POSSESSED);
    
#ifdef _WIN32
    // Windows 独自の設定
    WSADATA data;
    if(SOCKET_ERROR == WSAStartup(MAKEWORD(2, 0), &data)){
        CERR << "failed to initialize WSA-data." << endl;
        exit(1);
    }
#endif // _WIN32
    
    // open socket
    int sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0){
        CERR << "failed to open socket." << endl;
        exit(1);
    }
    //sockaddr_in 構造体のセット
    struct sockaddr_in addr;
    addr.sin_port = htons(port);
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr(host.c_str());
    // 接続
    if(connect(sock, (struct sockaddr*)&addr, sizeof(addr))){ // ソケット接続失敗
        CERR << "failed to connect to server." << endl;
        exit(1);
    }
    
    // 通信
    
    char rbuffer[65536];
    memset(rbuffer, 0, sizeof(rbuffer));
    std::string unfinished = "";
    
    while(1){
        if(strlen(rbuffer) <= 0){
            int bytes = recv(sock, rbuffer, sizeof(rbuffer), 0);
            CERR << bytes << endl;
            if(bytes == 0){
                CERR << "disconnected..." << endl;
                exit(1);
            }
        }
        std::string str = std::string(rbuffer);
        int len = strlen(rbuffer);
        DERR << len << endl;
        memmove(rbuffer, rbuffer + len + 1, sizeof(rbuffer) - len - 1);
        memset(rbuffer + sizeof(rbuffer) - len - 1, 0, len + 1);
        
        std::vector<std::string> commands = split(str, '\n');
        DERR << commands.size() << endl;
        
        for(std::string& command : commands){
            if(command.size() <= 0){ continue; }
            
            command = unfinished + command; // 前にjsonが完成していないテキストと繋げる
            CERR << "Server >> " << command << endl;
            
            // stringをjson形式に変換
            const char* json = command.c_str();
            picojson::value v;
            std::string jerr;
            picojson::parse(v, json, json + strlen(json), &jerr);
            if(jerr.empty()){
                picojson::object& o = v.get<picojson::object>();
                std::string req = o["request"].get<std::string>();
                //cerr << "request = " << req << endl;
                if(req == "NAME"){
                    if(sendMessage(sock, ai.name() + "\n") < 0){ exit(1); }
                }else if(req == "ROLE"){
                    CERR << ai.hopeRole() << endl;
                    if(sendMessage(sock, AIWolf::toRoleString(ai.hopeRole()) + "\n") < 0){ exit(1); }
                }else if(req == "INITIALIZE"){
                    updateInfoByJson(&ai.info(), o, AIWolf::GAME_INIT);
                    ai.initialize();
                    CERR << ai.info().toString();
                }else if(req == "DAILY_INITIALIZE"){
                    updateInfoByJson(&ai.info(), o, AIWolf::DAY_INIT);
                    // jsonからaiの局面情報を更新
                    ai.dailyInitialize();
                    CERR << ai.info().toString();
                }else if(req == "DAILY_FINISH"){
                    updateInfoByJson(&ai.info(), o, AIWolf::DAY_FIN);
                    // jsonからaiの局面情報を更新
                    ai.dailyFinish();
                    CERR << ai.info().toString();
                }else if(req == "FINISH"){
                    updateInfoByJson(&ai.info(), o, AIWolf::GAME_FIN);
                    // jsonからaiの局面情報を更新
                    ai.finish();
                    CERR << ai.info().toAllString();
                }else if(req == "TALK"){
                    updateInfoByJson(&ai.info(), o, AIWolf::GAME_FIN);
                    AIWolf::Talk talk = ai.talk();
                    CERR << ai.info().toAllString();
                    if(sendMessage(sock, AIWolf::toStatementString(talk) + "\n") < 0){ exit(1); }
                    if(talk != AIWolf::OVER){
                        //cerr << talk << endl; getchar();
                    }
                }else if(req == "WHISPER"){
                    updateInfoByJson(&ai.info(), o, AIWolf::GAME_FIN);
                    AIWolf::Whisper whisper = ai.whisper();
                    CERR << ai.info().toAllString();
                    if(sendMessage(sock, AIWolf::toStatementString(whisper) + "\n") < 0){ exit(1); }
                }else if(req == "VOTE"){
                    CERR << ai.info().toAllString();
                    if(sendMessage(sock, AIWolf::toAgentJsonString(ai.vote()) + "\n") < 0){ exit(1); }
                }else if(req == "DIVINE"){
                    CERR << ai.info().toAllString();
                    int t = ai.divine();
                    if(sendMessage(sock, AIWolf::toAgentJsonString(t) + "\n") < 0){ exit(1); }
                    if(ai.info().isValidDivine(ai.info().day(), ai.info().myIndex(), t)){
                        ai.info().setDivine(ai.info().day(), ai.info().myIndex(), t);
                    }else{
                        ai.info().setAbandonDivine(ai.info().day(), ai.info().myIndex());
                    }
                }else if(req == "GUARD"){
                    CERR << ai.info().toAllString();
                    int t = ai.guard();
                    if(sendMessage(sock, AIWolf::toAgentJsonString(t) + "\n") < 0){ exit(1); }
                    // guardはサーバーの受理or不受理が後の情報から判断できないので、ここで自分で記録
                    if(AIWolf::examAgentNum(t) && ai.info().isValidGuard(ai.info().day(), ai.info().myIndex(), t)){ // 合法なguard
                        ai.info().setGuard(ai.info().day(), ai.info().myIndex(), t);
                    }else{ // guard放棄
                        ai.info().setAbandonGuard(ai.info().day(), ai.info().myIndex());
                    }
                }else if(req == "ATTACK"){
                    CERR << ai.info().toAllString();
                    if(sendMessage(sock, AIWolf::toAgentJsonString(ai.attack()) + "\n") < 0){ exit(1); }
                }else{
                    CERR << "received unknown request" << endl;
                    exit(1);
                }
                unfinished = "";
            }else{
                // json形式が完成していない
                unfinished = command;
                CERR << jerr << endl;
            }
            CERR << "To Next Command" << endl;
            //getchar();
        }
        
        
        // engineに行動要求
        
        
        // サーバーに返答
        
    }
    
    // close socket
#ifdef _WIN32
    closesocket(sock);
    WSACleanup();
#else
    close(sock);
#endif // _WIN32
    return 0;
}
