/*
 aiwolf.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_JSON_HPP_
#define AIWOLF_JSON_HPP_

#include "./aiwolf.hpp"

#include "../../../lib/picojson.h"

namespace AIWolf{
    
    enum UpdateMode{
        INIT = 0,
        FIN,
        GAME_INIT,
        GAME_FIN,
        DAY_INIT,
        DAY_FIN,
    };
    
    /**************************盤面情報**************************/
    
#define JGet(o, s, t, X) if(!((o)[(s)].is<picojson::null>())){ t v = (o)[(s)].get<t>(); X; }
    
    template<class info_t>
    int updateTalkWhisperHistoryByJson(info_t *const pinfo, picojson::object& js){
        // 会話系は随時追加情報が来る毎回この関数で結果を取得
        if(!js["talkHistory"].is<picojson::null>()){
            picojson::array& talkHistory = js["talkHistory"].get<picojson::array>();
            if(talkHistory.size() > 0){
                int imax = 0;
                for(auto& jv : talkHistory){
                    picojson::object& jo = jv.get<picojson::object>();
                    int p = NONE_PLAYER, d = NONE_DAY, i = -1;
                    JGet(jo, "agent", double, p = toMyAgentNumber(int(v)); );
                    JGet(jo, "day", double, d = int(v); );
                    JGet(jo, "idx", double, i = int(v); imax = max(i, imax); );
                    JGet(jo, "content", std::string,
                         Talk t = toStatementData(v, TALK);
                         if(t == ILLEGAL_STATEMENT){
                             //cerr << "illegal talk! " << v << endl;// getchar();
                             t = SKIP;
                         }
                         pinfo->setPlayerTalk(d, i, p, t); );
                }
                pinfo->setTalkIndex(pinfo->day(), imax + 1);
            }
        }
        if(!js["whisperHistory"].is<picojson::null>()){
            picojson::array& whisperHistory = js["whisperHistory"].get<picojson::array>();
            if(whisperHistory.size() > 0){
                int imax = 0;
                for(auto& jv : whisperHistory){
                    picojson::object& jo = jv.get<picojson::object>();
                    int p = NONE_PLAYER, d = NONE_DAY, i = -1;
                    JGet(jo, "agent", double, p = toMyAgentNumber(int(v)); );
                    JGet(jo, "day", double, d = int(v); );
                    JGet(jo, "idx", double, i = int(v); imax = max(i, imax); );
                    JGet(jo, "content", std::string,
                         Whisper w = toStatementData(v, WHISPER);
                         if(w == ILLEGAL_STATEMENT){
                             //cerr << "illegal whisper! " << v << endl;// getchar();
                             w = SKIP;
                         }
                         pinfo->setPlayerWhisper(d, i, p, w); );
                }
                pinfo->setWhisperIndex(pinfo->day(), imax + 1);
            }
        }
        return 0;
    }
    
    template<class info_t>
    int updateInfoByJson(info_t *const pinfo, picojson::object& js, int mode = 0){
        /*if(js.find("day") != js.end()){ // 日付
            d = js["day"].get<int>();
            pinfo->setDay(d);
        }*/

        /*if(js.find("attackVoteList") != js.end()){
            int i = js["agent"].get<int>();
            pinfo->setMyIndex(toMyAgentNumber(i));
        }*/
        /*if(js.find("attackedAgent") != js.end()){
            int i = js["attackedAgent"].get<int>();
            pinfo->setAttacked(d, toMyAgentNumber(i));
        }
        if(js.find("executedAgent") != js.end()){
            int i = js["executedAgent"].get<int>();
            pinfo->setExecuted(d, toMyAgentNumber(i));
        }*/
        /*if(js.find("divineResult") != js.end()){
            int i = js["divineResult"].get<int>();
            pinfo->setDivined(toMyAgentNumber(i));
        }
        if(js.find("guardedAgent") != js.end()){
            int i = js["guardedAgent"].get<int>();
            pinfo->setGuardedAgent(toMyAgentNumber(i));
        }
        if(js.find("voteList") != js.end()){
            int i = js["voteList"].get<int>();
            pinfo->setGuardedAgent(toMyAgentNumber(i));
        }*/
        if(mode == GAME_INIT){
            
            // ルール系は試合中不変なので試合開始時のみ取得
            MatchRule rule;
            rule.clear();
            
            if(!js["gameSetting"].is<picojson::null>()){
                picojson::object& gameSetting = js["gameSetting"].get<picojson::object>();
                
                // プレーヤー人数
                JGet(gameSetting, "playerNum", double, rule.players = int(v); );
                // 会話系ターン数
                JGet(gameSetting, "maxTalk", double, rule.turns = int(v); );
                // 役職人数
                JGet(gameSetting, "roleNumMap", picojson::object,
                     picojson::object& roleNumMap = v;
                     iterateValidRole([&](int r){
                        JGet(roleNumMap, toRoleString(r), double, rule.rolePlayers.assign(r, int(v)));
                     }););
                
                pinfo->initMatch(rule);
            }
            
            pinfo->initGame();
            
            // 他の不変の情報
            if(!js["gameInfo"].is<picojson::null>()){
                picojson::object& gameInfo = js["gameInfo"].get<picojson::object>();
                
                // 自分の番号
                JGet(gameInfo, "agent", double,
                     pinfo->setMyIndex(toMyAgentNumber(int(v))); );
            }
            
        }else if(mode == DAY_INIT){
            if(!js["gameInfo"].is<picojson::null>()){
                picojson::object& gameInfo = js["gameInfo"].get<picojson::object>();
                // 主観的に判明している役職一覧
                JGet(gameInfo, "roleMap", picojson::object,
                     picojson::object& roleMap = v;
                     for(int i = 1; i <= pinfo->rule().players; ++i){
                         std::ostringstream oss;
                         oss << i;
                         JGet(roleMap, oss.str(), std::string,
                              pinfo->setRole(toMyAgentNumber(i), toRoleNum(v)); );
                     });
                // 日付
                JGet(gameInfo, "day", double, pinfo->setDay(int(v)); );
                // 襲撃結果
                JGet(gameInfo, "attackedAgent", double,
                     int p = int(v);
                     if(p < 0){ // 襲撃失敗または起こらなかった
                         pinfo->setNoAttacked(pinfo->day() - 1);
                     }else{
                         p = toMyAgentNumber(p);
                         pinfo->setAttacked(pinfo->day() - 1, p); // 前日
                         pinfo->setAttack(pinfo->day() - 1, p); // attack したことも確定
                     });
                // 処刑結果
                JGet(gameInfo, "executedAgent", double,
                     int p = int(v);
                     if(p < 0){ // 処刑執行されず
                         //pinfo->setNoExecuted(pinfo->day() - 1); // 現ルールでは無い
                     }else{
                         p = toMyAgentNumber(p);
                         pinfo->setExecuted(pinfo->day() - 1, p); // 前日
                     });
                // 占い結果
                JGet(gameInfo, "divineResult", picojson::object,
                     picojson::object& divineResult = v;
                     int p = NONE_PLAYER;
                     int d = NONE_DAY;
                     int r = NONE_SPECIES;
                     int t = NONE_PLAYER;
                     JGet(divineResult, "agent", double, p = toMyAgentNumber(int(v)); );
                     JGet(divineResult, "day", double, d = int(v); );
                     JGet(divineResult, "result", std::string, r = toSpeciesNum(v); );
                     JGet(divineResult, "target", double, t = toMyAgentNumber(int(v)); );
                     ASSERT(examAgentNum(t), cerr << "illegal divine target = " << t << endl;);
                     pinfo->setDivined(d - 1, p, t, r); );
                // 霊媒結果
                JGet(gameInfo, "mediumResult", picojson::object,
                     picojson::object& mediumResult = v;
                     int p = NONE_PLAYER;
                     int d = NONE_DAY;
                     int r = NONE_SPECIES;
                     int t = NONE_PLAYER;
                     JGet(mediumResult, "agent", double, p = toMyAgentNumber(int(v)); );
                     JGet(mediumResult, "day", double, d = int(v); );
                     JGet(mediumResult, "result", std::string, r = toSpeciesNum(v); );
                     JGet(mediumResult, "target", double, t = toMyAgentNumber(int(v)); );
                     ASSERT(examAgentNum(t), cerr << "illegal medium target = " << t << endl;);
                     pinfo->setInquested(d - 1, p, t, r); );
                // 処刑投票
                JGet(gameInfo, "voteList", picojson::array,
                     picojson::array& voteList = v;
                     for(int i = 0; i < (int)voteList.size(); ++i){
                         picojson::object& jo = voteList[i].get<picojson::object>();
                         int p = NONE_PLAYER;
                         int d = NONE_DAY;
                         int t = NONE_PLAYER;
                         JGet(jo, "agent", double, p = toMyAgentNumber(int(v)); );
                         JGet(jo, "day", double, d = int(v); );
                         JGet(jo, "target", double,
                              // 無効な投票はランダムに有効な投票に変えられる..と思っていたが実際に変な結果が来た
                              t = toMyAgentNumber(int(v)); );
                         //ASSERT(examAgentNum(t), cerr << "illegal vote target = " << t << endl;);
                         if(examAgentNum(t) && pinfo->isValidVote(d, p, t)){
                             pinfo->setVote(d, p, t);
                         }else{
                             pinfo->setAbstainVote(d, p);
                         }
                     }; );
                // 襲撃投票
                JGet(gameInfo, "attackVoteList", picojson::array,
                     picojson::array& attackVoteList = v;
                     for(int i = 0; i < (int)attackVoteList.size(); ++i){
                         picojson::object& jo = attackVoteList[i].get<picojson::object>();
                         int p = NONE_PLAYER;
                         int d = NONE_DAY;
                         int t = NONE_PLAYER;
                         JGet(jo, "agent", double, p = toMyAgentNumber(int(v)); );
                         JGet(jo, "day", double, d = int(v); );
                         JGet(jo, "target", double,
                              // 無効な投票が棄権となるので、そこのチェックが必須
                              t = int(v);
                              if(t < 0){
                                  t = NONE_PLAYER;
                              }else{
                                  t = toMyAgentNumber(t);
                              }; );
                         ASSERT(examAgentNum(t), cerr << "illegal attack-vote target = " << t << endl;);
                         if(t != NONE_PLAYER){
                             pinfo->setAttackVote(d, p, t);
                         }else{ //棄権
                             pinfo->setAbstainAttackVote(d, p);
                         }
                     }; );
                // プレーヤー状態
                /* 日付がわからないので使わない方がいいのでは
                 JGet(gameInfo, "statusMap", picojson::object,
                     picojson::object& statusMap = v;
                     for(int i = 0; i < N_MAX_PLAYERS; ++i){
                         int status = 0;
                         std::ostringstream oss;
                         oss << toOfficialAgentNumber(i);
                         JGet(statusMap, oss.str(), std::string,
                              if(v[0] == 'D'){ status = -1; }; );
                         if(status == 0){
                             pinfo->setAlive(d, i);
                         }
                     }; );*/
                pinfo->proceedDay(pinfo->day());
            }
        }else if(mode == DAY_FIN){
            updateTalkWhisperHistoryByJson(pinfo, js);
        }else if(mode == FIN){
        }else{
            // 行動決定前の情報通知
            updateTalkWhisperHistoryByJson(pinfo, js);
        }
        return 0;
    }
    
    /*template<class info_t>
    int updateInfoByJson(info_t *const pinfo, const picojson::object& js){
        if(js.find("agent") != js.end()){
            int i = js["agent"].get<int>();
            pinfo->setMyIndex(toMyAgentNumver(i));
        }
        if(js.find("attackVoteList") != js.end()){
            int i = js["agent"].get<int>();
            pinfo->setMyIndex(toMyAgentNumver(i));
        }
    }*/
    
#undef JGet
}

#endif // AIWOLF_JSON_HPP_