/*
 est_learner.cc
 Katsuki Ohto
 */

#include "../aiwolf.hpp"
#include "../thousand/singleEstimation.hpp"

int N = 10000;

using namespace AIWolf;
using namespace AIWolf::Thousand;

XorShift64 dice;

std::array<double, N_ROLES + 1> sameNumSum;
std::array<std::array<double, N_ROLES>, N_ROLES + 1> roleSameNumSum;
std::array<double, N_ROLES + 1> mdKLDSum;
std::array<int, N_ROLES + 1> kldSamples;
std::array<double, N_ROLES + 1> mcrSum;
std::array<double, N_ROLES + 1> bcrSum;
std::array<int, N_ROLES + 1> crSamples;

void initAyalysis(){
    sameNumSum.fill(0);
    for(int r = 0; r <= N_ROLES; ++r){
        roleSameNumSum[r].fill(0);
    }
    mdKLDSum.fill(0);
    kldSamples.fill(0);
    mcrSum.fill(0);
    bcrSum.fill(0);
    crSamples.fill(0);
}

template<class log_t, class estimator_t, class learner_t>
int learnEstParamsInAction(const log_t& log, const BitSet32 flags, const estimator_t& est, learner_t *const plearner){
    
    const int P = log.rule().players;
    
    double score[N_MAX_PLAYERS][N_ROLES] = {0};
    double tmpScore[N_MAX_PLAYERS][N_ROLES] = {0};
    
    ThinkInfo info(log);
    info.clearConversation();
    
    for(int d = 0; d < log.days(); ++d){
        info.proceedConversation(d); // 情報を更新
        info.setDay(d);
        
        // この時点で生きている全員の視点と、客観的視点で検証
        BitSet32 alive = info.alive(d).data();
        alive.set(31);
        iterate(alive, [&](int subject)->void{
            const int R = examAgentNum(subject) ? info.playerRole(subject) : NONE_ROLE;
            if(subject == 31){
                subject = -1;
            }else{
                info.sub[subject].setRole(subject, R, info); // 主観的な情報を設定
            }
            
            if(!flags.test(1) && dice.drand() > 0.5){ return; }
            
            // 全員の役職を推定
            for(int p = 0; p < P; ++p){
                
                auto rs = certainRoleOrSpecies(info, subject, p, R);
                if(rs.first != NONE_ROLE){
                    // subject視点で役職が確定する場合には学習は飛ばす
                    // ただし確率分布は計算しておく必要あり
                    for(int r = 0; r < N_ROLES; ++r){
                        score[p][r] = 0;
                    }
                    score[p][rs.first] = 1;
                    continue;
                }
                
                int idx = log.playerRole(p);
                
                switch(R){
                    case 0: calcEstimationScore<0, 1>(&score[p][0], info, subject, p, est); break;
                    case 1: calcEstimationScore<1, 1>(&score[p][0], info, subject, p, est); break;
                    case 2: calcEstimationScore<2, 1>(&score[p][0], info, subject, p, est); break;
                    case 3: calcEstimationScore<3, 1>(&score[p][0], info, subject, p, est); break;
                    case 4: calcEstimationScore<4, 1>(&score[p][0], info, subject, p, est); break;
                    case 5: calcEstimationScore<5, 1>(&score[p][0], info, subject, p, est); break;
                    default: calcEstimationScore<6, 0>(&score[p][0], info, -1, p, est); break;
                }
                
                if(rs.second != NONE_SPECIES){ // subject視点で種族が確定する場合には確率を調整
                    ASSERT(rs.second != Species::WEREWOLF,); // 人狼はここにこない
                    score[p][Role::WEREWOLF] = 0; // 人狼の確率0
                    ASSERT(info.playerRole(p) != Role::WEREWOLF, cerr << p << " by " << subject << endl << log.toString(););
                }
                
                if(subject >= 0){
                    if(flags.test(1)){ // feed feature value
                        plearner->feedFeatureValue(R);
                    }
                    if(flags.test(0)){ // learn
                        plearner->acceptDecidedMoveIndex(idx, R);
                        plearner->learnByMove();
                    }
                    if(flags.test(2)){ // test
                        plearner->feedObjValue(idx, R);
                    }
                }
            }
            if(flags.test(2) && !flags.test(0)){ // test onlyのとき
                
                /*for(int r = 0; r < N_ROLES; ++r){
                 for(int p = 0; p < P; ++p){
                 cerr << " " << score[p][r];
                 }cerr << endl;
                 }*/
                // 確実に知っている情報を追加
                //prepareWithCertainInfo(score,
                
                // 両方向で正規化し、正解とのKL情報量を計算
                memcpy(tmpScore, score, sizeof(tmpScore));
                normalize2(tmpScore, 8, log.rule());
                
                double kld = 0;
                for(int p = 0; p < P; ++p){
                    ASSERT(tmpScore[p][log.playerRole(p)] > 0, cerr << p << roleChar[log.playerRole(p)] << endl;);
                    kld += -std::log(tmpScore[p][log.playerRole(p)]);
                }
                mdKLDSum[R] += kld;
                kldSamples[R] += 1;
                
                /*for(int r = 0; r < N_ROLES; ++r){
                 for(int p = 0; p < P; ++p){
                 cerr << " " << int(score[p][r] * 1000);
                 }cerr << endl;
                 }
                 
                 for(int r = 0; r < N_ROLES; ++r){
                 for(int p = 0; p < P; ++p){
                 if(log.playerRole(p) == r){
                 cerr << " " << 1;
                 }else{
                 cerr << " " << 0;
                 }
                 }cerr << endl;
                 }*/
                
                // 正解率も再計算
                memcpy(tmpScore, score, sizeof(tmpScore));
                
                // 各役職の人数が等しいとして正規化
                normalize2<BASE_RULE.players, N_ROLES>(tmpScore, 8);
                // その結果一番のものが一致するか
                for(int p = 0; p < P; ++p){
                    auto v = getBestIndexVector(tmpScore[p], N_ROLES);
                    if(find(v.begin(), v.end(), log.playerRole(p)) != v.end()){
                        bcrSum[R] += 1;
                    }
                    crSamples[R] += 1;
                }
                
                // ランダム分配で何人当たるか調べる
                memcpy(tmpScore, score, sizeof(tmpScore));
                auto pr = dealRoleSubjective(tmpScore, info, subject, &dice);
                
                int same = countNCorrectRoles(pr, info.playerRoles(), P);
                
                sameNumSum[R] += same;
                
                for(int r = 0; r < N_ROLES; ++r){
                    int cnt = 0;
                    for(int p = 0; p < P; ++p){
                        if(pr[p] == r){ ++cnt; }
                    }
                    if(cnt != info.rule().rolePlayers[r]){
                        cerr << subject << roleChar[R];
                        cerr << " wrong number of role players " << toRoleString(r) << " : ";
                        cerr << cnt << " in " << info.rule().rolePlayers[r];
                        cerr << pr << info.playerRoles() << endl;
                        cerr << log.toString();
                        getchar();
                    }
                }
                
                for(int p = 0; p < P; ++p){
                    if(pr[p] == info.playerRole(p)){
                        roleSameNumSum[R][pr[p]] += 1 / double(info.rule().rolePlayers[pr[p]]);
                    }
                    if(R == Role::WEREWOLF && info.playerRole(p) == Role::WEREWOLF && pr[p] != Role::WEREWOLF){
                        // 人狼推定失敗
                        cerr << p << " by " << subject << endl << log.toString(); getchar();
                    }
                }
            }
        });
        info.addConversationActionRelation(d); // 宣言-行動関係を更新
    }
    return 0;
}

Estimator est;
EstimatorLearner learner;

int learn(const std::string& lpath, const std::string& opath, int mode, int iterations){
    
    double learnRate = (mode == 0) ? 1 : 0.8;
    
    std::mt19937 mt((uint32)time(NULL));
    
    est.setLearner(&learner);
    learner.setPolicy(&est);
    
    // ログをファイルから読み込み
    Log *const log = new Log[N];
    FILE *const pf = fopen(lpath.c_str(), "rb");
    fread(log, sizeof(Log) * N, 1, pf);
    fclose(pf);
    
    cerr << N << " games were imported." << endl;
    
    // 学習とテストの試合数決定
    const int games = N;
    const int learnGames = min((int)(games * learnRate), games);
    const int testGames = games - learnGames;
    
    // 完全ランダムの時の平均ヒット数をチェック
    {
        cerr << "base rule distribution = " << BASE_RULE.rolePlayers << endl; //getchar();
        double sum = 0;
        double pd[BASE_RULE.players][N_ROLES];
        double tmppd[BASE_RULE.players][N_ROLES];
        for(int p = 0; p < BASE_RULE.players; ++p){
            for(int r = 0; r < N_ROLES; ++r){
                pd[p][r] = BASE_RULE.rolePlayers[r] / (double) BASE_RULE.players;
            }
        }
        for(int i = 0; i < 10000; ++i){
            memcpy(tmppd, pd, sizeof(tmppd));
            auto a = dealRolePlain(tmppd, BASE_RULE, &dice);
            memcpy(tmppd, pd, sizeof(tmppd));
            auto b = dealRolePlain(tmppd, BASE_RULE, &dice);
            
            //cerr << a << b << endl;
            
            sum += countNCorrectRoles(a, b, BASE_RULE.players);
        }
        cerr << "hitNum at random = " << (sum / 10000) << endl;
    }
    
    // 特徴要素を解析して、学習の度合いを決める
    learner.setLearnParam(1, 0.00002, 0, 0.000002, 16);
    
    cerr << "started analyzing feature." << endl;
    
    learner.initFeatureValue();
    
    BitSet32 flag(0);
    flag.set(1);
    for(int i = 0; i < games; ++i){
        //cerr << log[i].toAllString(); getchar();
        //cerr << i << " ";
        learnEstParamsInAction(log[i], flag, est, &learner); // 特徴要素解析
    }
    learner.closeFeatureValue();
    learner.foutFeatureSurvey(opath + "est_feature_survey.dat");
    learner.finFeatureSurvey(opath + "est_feature_survey.dat");
    
    for(int r = 0; r <= N_ROLES; ++r){
        cerr << "record by " << roleChar[r] << " : " << learner.toRecordString(r) << endl;
    }
    
    std::vector<int> num;
    num.reserve(N);
    for(int i = 0; i < N; ++i){ num.push_back(i); }
    
    // 学習開始
    cerr << "started learning." << endl;
    
    for (int j = 0; j < iterations; ++j){
        
        cerr << "iteration " << j << endl;
        
        learner.setLearnParam(1, 0.00001 * (0.01 + 0.99 / sqrt(double(j + 1))), 0, 0.0000001 * (0.01 + 0.99 / sqrt(double(j + 1))), 256);
        
        BitSet32 flag(0);
        flag.set(0);
        flag.set(2);
        
        learner.initObjValue();
        std::shuffle(num.begin(), num.begin() + learnGames, mt); // 学習順入れ替え
        for(int i = 0; i < learnGames; ++i){
            learnEstParamsInAction(log[num[i]], flag, est, &learner); // 学習
        }
        est.hout("../c/param/est_param.h", "est_param");
        est.fout(opath + "est_param.dat");
        foutComment(est, opath + "est_comment.txt");
        for(int r = 0; r < N_ROLES; ++r){
            cerr << "Learning " << roleChar[r] << " : " << learner.toObjValueString(r) << endl;
        }
        
        if(testGames > 0){
            flag.reset(0);
            learner.initObjValue();
            
            initAyalysis();
            
            for(int i = learnGames; i < games; ++i){
                learnEstParamsInAction(log[i], flag, est, &learner); // テスト
            }
            for(int r = 0; r < N_ROLES; ++r){ // 全ての視点での結果を表示
                cerr << "Test " << roleChar[r] << " : " << learner.toObjValueString(r) << endl;
                cerr << "Test " << roleChar[r] << " : ";
                cerr << "normBHR = " << (bcrSum[r] / crSamples[r]) << " MD-KLD = " << (mdKLDSum[r] / kldSamples[r]);
                cerr << " hitNum = " << (sameNumSum[r] / kldSamples[r]) << endl;
            }
            cerr << "role hitNum = " << endl;
            for(int rr = 0; rr < N_ROLES; ++rr){
                cerr << " " << roleChar[rr] << " = ";
                for(int r = 0; r <= N_ROLES; ++r){
                    cerr << std::setw(5) << int(roleSameNumSum[r][rr] / kldSamples[r] * 1000);
                }cerr << endl;
            }
        }
    }
    cerr << "finished learning." << endl;
    
    return 0;
}

int main(int argc, char* argv[]){
    
    //基本的な初期化
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    
    std::string lpath = "/Users/ohto/documents/data/aiwolf/dump/binlog.bin";
    std::string opath = "/Users/ohto/documents/data/aiwolf/dump/";
    int mode = 0;
    int itr = 50;
    
    dice.srand((unsigned int)time(NULL) ^ 12345);
    
    for(int c = 1; c < argc; ++c){
        if(strstr(argv[c], "-l")){
            lpath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-n")){
            N = atoi(argv[c + 1]);
        }else if(strstr(argv[c], "-o")){
            opath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-t")){
            mode = 1;
        }
    }
    return learn(lpath, opath, mode, itr);
}