/*
 settings.h
 Katsuki Ohto
 */

#ifndef AIWOLF_SETTINGS_H_
#define AIWOLF_SETTINGS_H_

// thousand基本設定
// 設定パターンによっては動かなくなるかもしれないので注意

// プロフィール
#define MY_MC_NAME "Thousand"
#define MY_POL_NAME "Hundred"
#define MY_VERSION "20160306"
#define MY_COACH "KatsukiOhto"

// 重要な設定

#define MINIMUM // 本番用
//#define MONITOR // 着手決定関連の表示
//#define BROADCAST // 試合進行実況
//#define DEBUG // デバッグ出力。プレイアウトの内容も出力するので、重すぎて試合にならない。バグチェック用

constexpr int TIME_LIMIT_MS = 95; //　制限時間

// 試合に必要でないリッチ設定 本番用では自動オフ
// メソッド解析
#define USE_ANALYZER

#define CHECK_ALL_MOVES // 自分のプレイにて、必勝や諦めの判定がなされた後も生成された全ての着手を検討する

//#define POLICY_ONLY // 方策関数そのままでプレーする(末端探索は行う)
//#define FIXED_N_PLAYOUTS (5000) // プレイアウト回数を固定(デバッグ、実験用)

// 並列化
// スレッドの種類
using std::thread;

// スレッド数(ビルド時決定)
// 0以下を設定すると勝手に1になります
#define N_THREADS (1)

// ログ関連

// ログを取って外部ファイルに書き出す
#define LOGGING_FILE

// 置換表設定
#define USE_L2BOOK // ラスト2人置換表を使う
#define USE_LNCIBOOK // 3人以上の完全情報置換表を使う

// プレイアウト中の設定(学習時にも関係するので注意)
#define SEARCH_LEAF_MATE // プレイアウト末端で必勝探索を行う
#define SEARCH_LEAF_L2 // プレイアウト末端でラスト2人の全読みを行う
//#define SEARCH_LEAF_LNCI // プレイアウト末端でMaxN探索を行う

// 相手モデル解析設定
//#define MODELING_PLAY // 相手の着手をモデリング

/**************************以下は直接変更しない**************************/

//スレッド数
#ifdef N_THREADS
#if N_THREADS <= 0
#undef N_THREADS
#define N_THREADS (1)
#endif
#endif // N_THREADS

// 方策関数そのままのクライアントとしてのビルド
#ifdef POLICY_ONLY

// 名前
#define MY_NAME MY_POL_NAME

// スレッド数は1ということにする
// (0にすると探索用バッファが無くなる)
#ifdef N_THREADS
#undef N_THREADS
#define N_THREADS (1)
#endif

// ログ取りをするかどうかは、
// ファイル出力を行うかどうかに依存
#ifdef LOGGING_FILE
#ifndef LOGGING
#define LOGGING
#endif
#endif

#else // POLICY_ONLY

// モンテカルロありのとき

// 名前
#define MY_NAME MY_MC_NAME

// ログ取りは必ず行う
#ifndef LOGGING
#define LOGGING
#endif

#endif // POLICY_ONLY


// 本番用のとき、プレーに無関係なものは全て削除
// ただしバグ落ちにdefineが絡む可能性があるので、強制終了時の出力は消さない
#if defined(MINIMUM)

// デバッグ
#ifdef DEBUG
#undef DEBUG
#endif

// 実況
#ifdef BROADCAST
#undef BROADCAST
#endif

// アナライザ
#ifdef USE_ANALYZER
#undef USE_ANALYZER
#endif


#endif // MINIMUM

// スレッド数として2以上が指定された場合は、マルチスレッドフラグを立てる
#if N_THREADS >= (2)
#define MULTI_THREADING
#endif

#endif // AIWOLF_SETTINGS_H_
