/*
 thousand.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_HPP_
#define AIWOLF_THOUSAND_HPP_

#include "./aiwolf.hpp"
#include "./ai.hpp"

#include "./thousand/thinkInfo.hpp"

#include "./param/est_param.h"
#include "./param/eval_param.h"
#include "./param/pol_param.h"

#include "./thousand/singleEstimation.hpp"
#include "./thousand/evaluation.hpp"
#include "./thousand/actionPolicy.hpp"

#include "./thousand/playout.hpp"

namespace AIWolf{
    namespace Thousand{
        Estimator gEstimator;
        Evaluator gEvaluator;
        ActionPolicy gActionPolicy;
    }
}

#include "./thousand/action.hpp"

namespace AIWolf{
    class ThousandAI : public AI{
    private:
        XorShift64 dice_;
        bool actionDecided;
        BitArray32<16, 2> action; // 決定した行動を置いておく
        
        void decideDailyAction(){
            Thousand::ThinkInfo ti(info());
            ti.setFullConversation(ti.day());
            
            int myId = info().myIndex();
            
            switch(info().playerRole(myId)){
                case 0: action = Thousand::decideAction<0>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                case 1: action = Thousand::decideAction<1>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                case 2: action = Thousand::decideAction<2>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                case 3: action = Thousand::decideAction<3>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                case 4: action = Thousand::decideAction<4>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                case 5: action = Thousand::decideAction<5>(ti, myId,
                                                           Thousand::gEstimator,
                                                           Thousand::gEvaluator,
                                                           Thousand::gActionPolicy,
                                                           &dice_); break;
                default: break;
            }
            actionDecided = true;
        }
        
    public:
        virtual void initialize(){
            // パラメータ読み込み(試合レギュレーションによりコードの一部として置いている)
            Thousand::gEstimator.setParam(est_param);
            Thousand::gEvaluator.setParam(eval_param);
            Thousand::gActionPolicy.setParam(pol_param);
            
            // サイコロ初期化
            std::random_device seed;
            dice_.srand(seed() ^ (unsigned int)time(NULL));
        }
        
        virtual void dailyInitialize(){
            actionDecided = false;
        }
        
        virtual void dailyFinish(){
        }
        
        virtual void finish(){
        }
        
        virtual Talk talk(){
            Thousand::ThinkInfo ti(info());
            ti.setFullConversation(ti.day());
            
            Talk t = OVER;
            int myId = info().myIndex();
            
            switch(info().playerRole(myId)){
                case 0: t = Thousand::talk<0>(ti, myId, Thousand::gEstimator, &dice_); break;
                case 1: t = Thousand::talk<1>(ti, myId, Thousand::gEstimator, &dice_); break;
                case 2: t = Thousand::talk<2>(ti, myId, Thousand::gEstimator, &dice_); break;
                case 3: t = Thousand::talk<3>(ti, myId, Thousand::gEstimator, &dice_); break;
                case 4: t = Thousand::talk<4>(ti, myId, Thousand::gEstimator, &dice_); break;
                case 5: t = Thousand::talk<5>(ti, myId, Thousand::gEstimator, &dice_); break;
                default: break;
            }
            ASSERT(info().isValidTalk(info().day(), myId, t),
                   cerr << "Thousand::vote() : illegal talk" << t << endl << info().toString(););
            return t;
        }
        
        virtual Whisper whisper(){
            return OVER;
        }
        
        virtual int vote(){
            int v = -1;
            if(!actionDecided){ decideDailyAction(); }
            v = action[0]; // 投票先
            ASSERT(info().isValidVote(info().day(), info().myIndex(), v),
                   cerr << "Thousand::vote() : illegal vote " << v << endl << info().toString(););
            return v;
        }
        
        virtual int divine(){
            int d = -1;
            if(!actionDecided){ decideDailyAction(); }
            d = action[1]; // 占い先
            ASSERT(info().isValidDivine(info().day(), info().myIndex(), d),
                   cerr << "Thousand::divine() : illegal divine " << d << endl << info().toString(););
            return d;
        }
        
        virtual int guard(){
            int g = -1;
            if(!actionDecided){ decideDailyAction(); }
            g = action[1]; // 護衛先
            ASSERT(info().isValidGuard(info().day(), info().myIndex(), g),
                   cerr << "Thousand::guard() : illegal guard " << g << endl << info().toString(););
            return g;
        }
        
        virtual int attack(){
            int a = -1;
            if(!actionDecided){ decideDailyAction(); }
            a = action[1]; // 投票先
            ASSERT(info().isValidAttackVote(info().day(), info().myIndex(), a),
                   cerr << "Thousand::attack() : illegal attack " << a << endl << info().toString(););
            return a;
        }
    };
}

#endif // AIWOLF_THOUSAND_HPP_