/*
 action.hpp
 Katsuki Ohto
*/

#ifndef AIWOLF_THOUSAND_ACTION_HPP_
#define AIWOLF_THOUSAND_ACTION_HPP_

#include "../aiwolf.hpp"

#include "./thinkInfo.hpp"

#include "./evaluation.hpp"
#include "./singleEstimation.hpp"

namespace AIWolf{
    namespace Thousand{
        
        template<int R, class estimator_t, class dice_t>
        int talk(ThinkInfo& info, int player,
                 const estimator_t& est,
                 dice_t *const pdice){
            // 表の会話
            // とりあえずほとんど何もしない
            const int day = info.day();
            const int turn = info.talkTurn();
            
            info.sub[player].setRole(player, R, info);
            estimateNow(&info.sub[player], player, info, est);
            
            const int P = info.rule().players;
            
            Talk t;
            
            static int e = 0; // 推定発言をしたか
            static int v = 0; // 投票先発言をしたか
            if(turn == 0){ e = 0; v = 0; }
            
            
            if(R == Role::SEER){ // 占い師のとき
                if(day == DAY_DIVINE_FIRST){
                    if(turn == 0){ // 占い師CO
                        t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::SEER); return t;
                    }else{
                        return OVER;
                    }
                }else{
                    if(turn == 0){ // 占い結果報告
                        t.setType(DIVINED); t.setDst(info.divine(day - 1)); t.setRole(info.divined(day - 1)); return t;
                    }else{
                        return OVER;
                    }
                }
            }else if(R == Role::POSSESSED){
                if(day == DAY_DIVINE_FIRST){
                    if(turn == 0){ // 占い師騙り
                        t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::SEER); return t;
                    }else{
                        return OVER;
                    }
                }else{
                    if(turn == 0){ // 占い結果の嘘報告
                        auto alive = info.alive(day - 1); // 昨日の時点で生きているプレーヤーの全列挙
                        alive.reset(player);
                        for(int p = 0; p < P; ++p){
                            if(info.divineResult[player][p] != NONE_SPECIES){
                                alive.reset(p);
                            }
                        }
                        if(alive.any()){
                            // 最も人狼率が高いプレーヤーを人間判定し、
                            // 最も人間率が高いプレーヤーを人狼判定する
                            int w = 0;
                            double wsc = -9999;
                            int h = 0;
                            double hsc = +9999;
                            double hsum = 0;
                            double wsum = 0;
                            
                            for(int p = 0; p < P; ++p){
                                double sc = info.sub[player].estPD[p][Role::WEREWOLF];
                                if(alive.test(p)){
                                    if(sc > wsc){
                                        wsc = sc;
                                        w = p;
                                    }
                                    if(sc < hsc){
                                        hsc = sc;
                                        h = p;
                                    }
                                }
                                if(info.alive(day - 1)){
                                    hsum += 1 - sc;
                                    wsum += sc;
                                }
                            }
                            //cerr << alive << endl;
                            //cerr << "hsum = " << hsum << " wsum = " << wsum << endl;
                            //cerr << "best h = " << h << " best w = " << w << endl; getchar();
                            double dr = pdice->drand() * (hsum + wsum);
                            if(dr > hsum){
                                t.setType(DIVINED); t.setDst(w); t.setRole(Species::HUMAN); return t;
                            }else{
                                t.setType(DIVINED); t.setDst(h); t.setRole(Species::WEREWOLF); return t;
                            }
                        }
                    }
                }
            }else if(R == Role::MEDIUM){
                if(day > DAY_EXECUTE_FIRST){
                    if(info.comingOutRole[player] != Role::MEDIUM){
                        if(info.inquested(day - 1) == Species::WEREWOLF){ // 人狼発見
                            if(turn == 0){ // 霊媒師CO
                                t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::MEDIUM); return t;
                            }
                        }
                    }else{ // すでにCOした
                        // これまで全ての霊媒結果報告
                        for(int d = DAY_EXECUTE_FIRST; d < day; ++d){
                            if(info.inquestResult[player][info.executed(d)] == NONE_SPECIES){ // 霊媒結果をまだ言っていない
                                t.setType(INQUESTED); t.setDst(info.executed(d)); t.setRole(info.inquested(d)); return t;
                            }
                        }
                    }
                }
            }
            return OVER;
        }
        /*
        template<class estimator_t, class dice_t>
        int whisper(ThinkInfo& info, int player,
                 const estimator_t& est,
                 dice_t *const pdice){
            // 表の会話
            // とりあえずほとんど何もしない
            const int day = info.day();
            const int turn = info.talkTurn();
            
            info.sub[player].setRole(player, R, info);
            estimateNow(&info.sub[player], player, info, est);
            
            const int P = info.rule().players;
            
            Talk t;
            
            if(R == Role::SEER){ // 占い師のとき
                if(day == DAY_DIVINE_FIRST){
                    if(turn == 0){ // 占い師CO
                        t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::SEER); return t;
                    }else{
                        return OVER;
                    }
                }else{
                    if(turn == 0){ // 占い結果報告
                        t.setType(DIVINED); t.setDst(info.divine(day - 1)); t.setRole(info.divined(day - 1)); return t;
                    }else{
                        return OVER;
                    }
                }
            }else if(R == Role::POSSESSED){
                if(day == DAY_DIVINE_FIRST){
                    if(turn == 0){ // 占い師騙り
                        t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::SEER); return t;
                    }else{
                        return OVER;
                    }
                }else{
                    if(turn == 0){ // 占い結果の嘘報告
                        auto alive = info.alive(day - 1); // 昨日の時点で生きているプレーヤーの全列挙
                        alive.reset(player);
                        for(int p = 0; p < P; ++p){
                            if(info.divineResult[player][p] != NONE_SPECIES){
                                alive.reset(p);
                            }
                        }
                        if(alive.any()){
                            // 最も人狼率が高いプレーヤーを人間判定し、
                            // 最も人間率が高いプレーヤーを人狼判定する
                            int w = 0;
                            double wsc = -9999;
                            int h = 0;
                            double hsc = +9999;
                            double hsum = 0;
                            double wsum = 0;
                            
                            for(int p = 0; p < P; ++p){
                                double sc = info.sub[player].estPD[p][Role::WEREWOLF];
                                if(alive.test(p)){
                                    if(sc > wsc){
                                        wsc = sc;
                                        w = p;
                                    }
                                    if(sc < hsc){
                                        hsc = sc;
                                        h = p;
                                    }
                                }
                                if(info.alive(day - 1)){
                                    hsum += 1 - sc;
                                    wsum += sc;
                                }
                            }
                            //cerr << alive << endl;
                            //cerr << "hsum = " << hsum << " wsum = " << wsum << endl;
                            //cerr << "best h = " << h << " best w = " << w << endl; getchar();
                            double dr = pdice->drand() * (hsum + wsum);
                            if(dr > hsum){
                                t.setType(DIVINED); t.setDst(w); t.setRole(Species::HUMAN); return t;
                            }else{
                                t.setType(DIVINED); t.setDst(h); t.setRole(Species::WEREWOLF); return t;
                            }
                        }
                    }
                }
            }else if(R == Role::MEDIUM){
                if(day > DAY_EXECUTE_FIRST){
                    if(info.comingOutRole[player] != Role::MEDIUM){
                        if(info.inquested(day - 1) == Species::WEREWOLF){ // 人狼発見
                            if(turn == 0){ // 霊媒師CO
                                t.setType(COMINGOUT); t.setDst(player); t.setRole(Role::MEDIUM); return t;
                            }
                        }
                    }else{ // すでにCOした
                        // これまで全ての霊媒結果報告
                        for(int d = DAY_EXECUTE_FIRST; d < day; ++d){
                            if(info.inquestResult[player][info.executed(d)] == NONE_SPECIES){ // 霊媒結果をまだ言っていない
                                t.setType(INQUESTED); t.setDst(info.executed(d)); t.setRole(info.inquested(d)); return t;
                            }
                        }
                    }
                }
            }
            return OVER;
        }*/
        
        template<int R, class estimator_t, class evaluator_t, class policy_t, class dice_t>
        BitArray32<16, 2> decideAction(ThinkInfo& info, int player,
                                       const estimator_t& est,
                                       const evaluator_t& eval,
                                       const policy_t& pol,
                                       dice_t *const pdice){
            // 投票
            ASSERT(info.exam(), cerr << info.toString() << endl;);
            
            ClockMS clock;
            clock.start();
            
            const int P = info.rule().players;
            
            info.sub[player].setRole(player, R, info);
            estimateNow(&info.sub[player], player, info, est);
            
            // 全合法行動(まともなもののみ)の組み合わせを列挙
            constexpr int N_MAX_MOVES = N_MAX_PLAYERS * N_MAX_PLAYERS;
            BitArray32<16, 2> mv[N_MAX_MOVES];
            const int NMoves = genValidActions(mv, info, player); // 行動生成
            
            //cerr << NMoves << " moves were generated." << endl;
            
            ASSERT(NMoves > 0, cerr << NMoves << endl;);
            
#if 0
            double pd[N_MAX_PLAYERS][N_ROLES] = {0};
            // 主観役職推定
            calcEstimationScoreAll<R, 0>(pd, info, player, est);
            
            // 確実な主観情報の利用
            auto roleNum = info.rule().rolePlayers;
            BitArray64<4, 16> ret(0);
            auto seq = prepareWithCertainInfo<R>(pd, ret, roleNum, info, player);
            
            // 役職決定していないプレーヤーを正規化
            normalize2(pd, 8, roleNum.sum(), seq, roleNum);
            
            //cerr << pd;
            
            //
            //memcpy(tpd, pd, sizeof(tpd));
            //auto pr = dealRole<R>(tpd, info, player, pdice);
        
            double sum[N_MAX_MOVES] = {0};
            std::array<uint32, N_MAX_MOVES> size;
            uint32 sizeSum = 2 * NMoves;
            int trials = 0;
            
            size.fill(2);
     

            // モンテカルロで行動決定
            while(clock.stop() < TIME_LIMIT_MS){ // 制限時間一杯まで考える
                // どの着手をトライするか決定
                int index = 0;
                double bestUCB = -9999;
                
                for(int m = 0; m < NMoves; ++m){
                    double ucb = calc_ucb1(sum[m] / size[m], size[m], sizeSum, 2.0);
                    if(ucb > bestUCB){
                        index = m;
                        bestUCB = ucb;
                    }
                }
                
                // 局面を生成
                double tpd[N_MAX_PLAYERS][N_ROLES];
                memcpy(tpd, pd, sizeof(tpd));
                auto pr = dealRole<R>(tpd, info, player, pdice);
                
                // シミュレーションを行う
                double ev = doSimulation(info, pr, player, mv[index], est, eval, pol, pdice);
                
                if(ev < -10){ continue; }// 無効な局面であった
                
                CERR << "playout " << trials << " : " << mv[index] << pr << ev << endl; //getchar();
                
                if(getTeam(R) == Team::WEREWOLF){ ev = -ev; } // 人狼サイドでは評価を逆転
                
                // 値を更新
                sum[index] += ev;
                size[index] += 1;
                sizeSum += 1;
                ++trials;
            }
            
            // 思考終了して最も高評価な行動に決定
            int best = 0;
            double bestEv = -9999;
            for(int m = 0; m < NMoves; ++m){
                double mean = sum[m] / size[m];
                if(mean > bestEv){
                    best = m;
                    bestEv = mean;
                }
                //cerr << mv[m] << " " << mean << endl;
            }
            
            //cerr << trials << " playouts were done." << endl;
            //cerr << mv[best] << endl;
            return mv[best];
#endif
            // 行動価値関数一手読みで行動決定
            
            //cerr << info.sub[player].estPD;
            //getchar();
            
            double score[N_MAX_MOVES];
            double sum = calcActionPolicyScore<R, 0>(score, mv, NMoves, info, player, pol);
            
            int bestCount = 2;
            int best = 0;
            double bestScore = -9999;
            // 温度0で選択
            for(int m = 0; m < NMoves; ++m){
                //cerr << mv[m] << score[m] << endl;
                if(score[m] > bestScore){
                    best = m; bestScore = score[m]; bestCount = 2;
                }else if(score[m] == bestScore){
                    if(pdice->rand() % bestCount == 0){
                        best = m;
                    }
                    ++bestCount;
                }
            }
            //cerr << mv[best] << " were chosen." << endl;
            
            return mv[best];
        }
    }
}

#endif // AIWOLF_THOUSAND_PLAYOUT_HPP_
