/*
 actionPolicy.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_ACTIONPOLICY_HPP_
#define AIWOLF_THOUSAND_ACTIONPOLICY_HPP_

#include "../aiwolf.hpp"
#include "./thinkInfo.hpp"

namespace AIWolf{

    /**************************行動方策**************************/
    
    namespace ActionPolicySpace{
        enum{
            
            // 処刑投票(全員)
            FEA_VOTE_EQUAL_DEC = 0, // 宣言通りの相手に投票する
            FEA_VOTE_ROLE_PROB, // 主観的な役職確率と2乗和
            FEA_VOTE_ROLE_EST, // 推定宣言の利用
            FEA_VOTE_HIGHEST_VOTED_DEC, // 自分以外で最多得票だったプレーヤーに投票
            FEA_VOTE_EQUAL_YESTERDAY, // 前日と同じプレーヤーに投票
            
            // 処刑投票(村人陣営)
            FEA_VOTE_SEER, // 占い師がCOしているとき占い師CO者に投票するか
            FEA_VOTE_SEER_PROB_DIVINED, // 占い師CO者の結果を信頼して投票
            
            // 処刑投票(狂人)
            
            // 処刑投票(人狼陣営)
            FEA_VOTE_WEREWOLF_EQUAL_DEC, // 裏の宣言通りの相手に投票する
            FEA_VOTE_WEREWOLF_OTHER_DEC, // 他の人狼の裏の投票宣言先に合わせる
            FEA_VOTE_WEREWOLF_AGR, // 他の人狼の裏の投票先で合意が取れている場合
            FEA_VOTE_WEREWOLF_OTHER_AGR, // 他の人狼同士で合意が取れており自分だけ外れている場合
            
            // 占い
            FEA_DIVINE_EQUAL_VOTE, // 投票先と同じ相手を占う
            FEA_DIVINE_ROLE_PROB, // 主観的な役職確率と2乗和
            
            // 護衛
            FEA_GUARD_EQUAL_VOTE, // 投票先と同じ相手を護衛
            FEA_GUARD_ROLE_PROB, // 主観的な役職確率と2乗和
            
            // 襲撃
            FEA_ATTACK_VOTE_EQUAL_DEC, // 宣言通りの相手に襲撃投票する
            FEA_ATTACK_VOTE_EQUAL_VOTE, // 投票先と同じ相手に襲撃投票
            FEA_ATTACK_VOTE_OTHER_DEC, // 他の人狼の裏の投票宣言先に合わせる
            FEA_ATTACK_VOTE_AGR, // 他の人狼の裏の投票先で合意が取れている場合
            FEA_ATTACK_VOTE_OTHER_AGR, // 他の人狼同士で合意が取れており自分だけ外れている場合
            FEA_ATTACK_VOTE_ROLE_PROB, // 主観的な役職確率と2乗和
            FEA_ATTACK_VOTE_EQUAL_YESTERDAY, // 前日と同じプレーヤーに襲撃投票
            FEA_ATTACK_VOTE_SEER, // 占い師がCOしているとき占い師CO者を襲撃するか
            
            FEA_ALL
        };
        
        constexpr int numTable[] = {
            N_ROLES * 2,
            N_ROLES * N_ROLES * 2,
            0,
            0,
            N_ROLES,
            
            N_ROLES * 2,
            2,
            
            2,
            0,
            0,
            0,
            
            1,
            N_ROLES * 2,
            
            1,
            N_ROLES * 2,
            
            2,
            1,
            0,
            0,
            0,
            N_ROLES * 2,
            1,
            2,
        };
        static_assert(FEA_ALL == sizeof(numTable) / sizeof(int), "num of features confliction.");
        
        constexpr int FEA_NUM(unsigned int fea){
            return numTable[fea];
        }
        constexpr int FEA_IDX(unsigned int fea){
            return (fea == 0) ? 0 : (FEA_IDX(fea - 1) + FEA_NUM(fea - 1));
        }
        constexpr int FEA_NUM_ALL = FEA_IDX(FEA_ALL);
 
#define LINEOUT(fea, str) { out << str << endl; int base = FEA_IDX(fea); for (int i = 0; i < FEA_NUM(fea); ++i){os(base + i); } out << endl; }
        
#define LINEOUTX(fea, str, x) { out << str << endl; int base = FEA_IDX(fea); int num = FEA_NUM(fea);\
for (int i = 0;;){os(base + i); ++i; if(i >= num)break; if(i % (x) == 0){ out << endl; }} out << endl; }
        
        int commentParam(std::ostream& out, const double param[FEA_NUM_ALL]){
            auto os = [&out, param](int idx)->void{ out << param[idx] << " "; };

            out << "****** POLICY ******" << endl;
            
            LINEOUTX(FEA_VOTE_EQUAL_DEC, "general - vote = vdec", N_ROLES);
            LINEOUTX(FEA_VOTE_ROLE_PROB, "vote - est role prob", N_ROLES);
            LINEOUT(FEA_VOTE_ROLE_EST, "role by est dec");
            LINEOUT(FEA_VOTE_HIGHEST_VOTED_DEC, "");
            LINEOUTX(FEA_VOTE_EQUAL_YESTERDAY, "vote = yesterday vote", N_ROLES);
            
            // 処刑投票(村人陣営)
            LINEOUTX(FEA_VOTE_SEER, "vote seer CO", N_ROLES);
            LINEOUT(FEA_VOTE_SEER_PROB_DIVINED, "vote using seer-divine result");
            
            // 処刑投票(人狼陣営)
            LINEOUT(FEA_VOTE_WEREWOLF_EQUAL_DEC, "whisper vote = adec");
            LINEOUT(FEA_VOTE_WEREWOLF_OTHER_DEC, "");
            LINEOUT(FEA_VOTE_WEREWOLF_AGR, "");
            LINEOUT(FEA_VOTE_WEREWOLF_OTHER_AGR, "");
            
            // 占い
            LINEOUT(FEA_DIVINE_EQUAL_VOTE, "divine equal vote");
            LINEOUTX(FEA_DIVINE_ROLE_PROB, "divine - est role prob", N_ROLES);
            
            // 護衛
            LINEOUT(FEA_GUARD_EQUAL_VOTE, "guard equal vote");
            LINEOUTX(FEA_GUARD_ROLE_PROB, "guard - est role prob", N_ROLES);
            
            // 襲撃
            LINEOUT(FEA_ATTACK_VOTE_EQUAL_DEC, "attack vote = adec");
            LINEOUT(FEA_ATTACK_VOTE_EQUAL_VOTE, "");
            LINEOUT(FEA_ATTACK_VOTE_OTHER_DEC, "");
            LINEOUT(FEA_ATTACK_VOTE_AGR, "");
            LINEOUT(FEA_ATTACK_VOTE_OTHER_AGR, "");
            LINEOUTX(FEA_ATTACK_VOTE_ROLE_PROB, "attack - est role prob", N_ROLES);
            LINEOUT(FEA_ATTACK_VOTE_EQUAL_YESTERDAY, "avote = yesterday avote");
            LINEOUT(FEA_ATTACK_VOTE_SEER, "attack seer CO");
            
            return 0;
        }
#undef LINEOUT
    }
    
    using ActionPolicy = SoftmaxPolicy<ActionPolicySpace::FEA_NUM_ALL, N_ROLES>;
    using ActionPolicyLearner = SoftmaxPolicyLearner<ActionPolicy>;
    
    int foutComment(const ActionPolicy& pol, const std::string& fName){
        std::ofstream ofs(fName, std::ios::out);
        return ActionPolicySpace::commentParam(ofs, pol.param_);
    }
    
#define Foo(i) {s += pol.param(i); pol.template feedFeatureScore<M>((i), 1.0);}
#define FooX(i, x) s += pol.param(i) * (x); pol.template feedFeatureScore<M>((i), (x));}
    
#define FooC(p, i) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
    cerr << "FooC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
    cerr << " (pi = " << (p) << ")" << endl; );\
    s += pol.param(i); pol.template feedFeatureScore<M>((i), 1.0);}
    
#define FooXC(p, i, x) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
    cerr << "FooXC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
    cerr << " (pi = " << (p) << ")" << endl; );\
    s += pol.param(i) * (x); pol.template feedFeatureScore<M>((i), (x));}
    
    double rate(int a, int b){
        return (b == 0) ? 0.5 : (a / (double)b);
    }
    double logit(double x, double clip = 10){
        
    }

    template<int R, int M = 1, class move_t, class info_t, class policy_t>
    double calcActionPolicyScore(double *const dst, const move_t *const mv, const int NMoves,
                         const info_t& info, const int player, const policy_t& pol){
    
        using namespace ActionPolicySpace;
        
        pol.template initCalculatingScore<M>(NMoves);
        
        const unsigned int day = info.day();
        const int P = info.rule().players;
        
        // 推定確率に対する前処理
        //double l[N_ROLES][2];
        //for(int r = 0; r < N_ROLES; ++r){
         //   double l = logit(info.sub[player].estPD[d0][r]);
        double sum = 0;
        
        for(int m = 0; m < NMoves; ++m){
            
            pol.template initCalculatingCandidateScore<M>();
            
            int d0 = mv[m][0], d1 = mv[m][1];
            
            ASSERT(0 <= d0 && d0 < info.rule().players, cerr << d0 << endl;);
            ASSERT(0 <= d1 && d1 < info.rule().players, cerr << d1 << endl;);
            
            double s = 0;
            
            if(day >= DAY_EXECUTE_FIRST){
                // 宣言通りの相手に投票する
                /*{
                    constexpr int base = FEA_IDX(FEA_VOTE_EQUAL_DEC);
                    if(d0 == info.voteDec[day][player]){ // 自分に対する投票は生成されていないので確認しなくて良い
                        FooC(FEA_VOTE_EQUAL_DEC, base + R);
                        FooXC(FEA_VOTE_EQUAL_DEC, base + N_ROLES + R,
                              rate(info.voteEqualDecNum, info.voteDecDayNum));
                    }
                }*/
                
                // 主観役職確率と処刑投票
                {
                    constexpr int base = FEA_IDX(FEA_VOTE_ROLE_PROB);
                    for(int r = 0; r < N_ROLES; ++r){
                        double l = info.sub[player].estPD[d0][r] - 0.5; //logit(info.sub[player].estPD[d0][r]);
                        FooXC(FEA_VOTE_ROLE_PROB, base + R * N_ROLES + r, l);
                        FooXC(FEA_VOTE_ROLE_PROB, base + N_ROLES * N_ROLES + R * N_ROLES + r, pow(l, 2));
                    }
                }
                // 占い師CO者に投票する
                {
                    constexpr int base = FEA_IDX(FEA_VOTE_SEER);
                    if(day >= DAY_EXECUTE_FIRST){
                        if(info.comingOutRole[d0] == Role::SEER){ // 占い師CO者
                            int coNum = info.comingOutedNum[Role::SEER];
                            if(coNum > 1){ // 複数CO
                                FooC(FEA_VOTE_SEER, base + N_ROLES + R);
                            }else{
                                FooC(FEA_VOTE_SEER, base + R);
                            }
                        }
                    }
                }
                // 占い師CO者の結果を信頼して投票する
                if(R != Role::SEER){
                    constexpr int base = FEA_IDX(FEA_VOTE_SEER_PROB_DIVINED);
                    if(day >= DAY_EXECUTE_FIRST){
                        double rate = 0;
                        // それぞれの占い師CO者の出している結果を総計
                        int cnt = 0;
                        for(int p = 0; p < P; ++p){
                            if(info.comingOutRole[p] == Role::SEER){ // 占い師CO者
                                if(info.divineResult[p][d0] == Species::WEREWOLF){
                                    rate -= info.sub[player].estPD[p][Role::SEER];
                                    ++cnt;
                                }else if(info.divineResult[p][d0] == Species::HUMAN){
                                    rate += info.sub[player].estPD[p][Role::SEER];
                                    ++cnt;
                                }
                            }
                        }
                        if(cnt > 0){
                            rate /= cnt;
                            if(R == Role::WEREWOLF){ // 複数CO
                                FooXC(FEA_VOTE_SEER_PROB_DIVINED, base + 1, rate);
                            }else{
                                FooXC(FEA_VOTE_SEER_PROB_DIVINED, base, rate);
                            }
                        }
                    }
                }
            }
            
            // 前日と同じ相手に投票する
            /*{
                constexpr int base = FEA_IDX(FEA_VOTE_EQUAL_YESTERDAY);
                if(day >= DAY_EXECUTE_FIRST + 1){
                    if(d0 == info.vote(day - 1, player)){
                        FooC(FEA_VOTE_EQUAL_YESTERDAY, base + R);
                    }
                }
            }*/
            // <各役職の行動>
            
            // 投票と同じ相手を占う
            if(R == Role::SEER){
                constexpr int base = FEA_IDX(FEA_DIVINE_EQUAL_VOTE);
                if(day >= DAY_EXECUTE_FIRST && day >= DAY_DIVINE_FIRST){
                    if(d0 == d1){ // 同じ
                        FooC(FEA_DIVINE_EQUAL_VOTE, base);
                    }
                }
            }
            // 主観役職確率と占い
            if(R == Role::SEER){
                constexpr int base = FEA_IDX(FEA_DIVINE_ROLE_PROB);
                if(day >= DAY_DIVINE_FIRST){
                    for(int r = 0; r < N_ROLES; ++r){
                        double l = info.sub[player].estPD[d1][r] - 0.5;
                        FooXC(FEA_DIVINE_ROLE_PROB, base + r, l);
                        FooXC(FEA_DIVINE_ROLE_PROB, base + N_ROLES + r, pow(l, 2));
                    }
                }
            }
            
            // 投票と同じ相手を護衛
            if(R == Role::BODYGUARD){
                constexpr int base = FEA_IDX(FEA_GUARD_EQUAL_VOTE);
                if(day >= DAY_EXECUTE_FIRST && day >= DAY_ATTACK_FIRST){
                    if(d0 == d1){ // 同じ
                        FooC(FEA_GUARD_EQUAL_VOTE, base);
                    }
                }
            }
            // 主観役職確率と護衛
            if(R == Role::BODYGUARD){
                constexpr int base = FEA_IDX(FEA_GUARD_ROLE_PROB);
                if(day >= DAY_ATTACK_FIRST){
                    for(int r = 0; r < N_ROLES; ++r){
                        double l = info.sub[player].estPD[d1][r] - 0.5;
                        FooXC(FEA_GUARD_ROLE_PROB, base + r, l);
                        FooXC(FEA_GUARD_ROLE_PROB, base + N_ROLES + r, pow(l, 2));
                    }
                }
            }
            
            // 処刑投票と同じ相手を襲撃
            if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_ATTACK_VOTE_EQUAL_VOTE);
                if(day >= DAY_EXECUTE_FIRST && day >= DAY_ATTACK_FIRST){
                    if(d0 == d1){ // 同じ
                        FooC(FEA_ATTACK_VOTE_EQUAL_VOTE, base);
                    }
                }
            }
            // 主観役職確率と襲撃
            if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_ATTACK_VOTE_ROLE_PROB);
                if(day >= DAY_ATTACK_FIRST){
                    for(int r = 0; r < N_ROLES; ++r){
                        double l = info.sub[player].estPD[d1][r] - 0.5;
                        FooXC(FEA_ATTACK_VOTE_ROLE_PROB, base + r, l);
                        FooXC(FEA_ATTACK_VOTE_ROLE_PROB, base + N_ROLES + r, pow(l, 2));
                    }
                }
            }
            
            // 裏での宣言通りの相手を投票
            /*if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_VOTE_WEREWOLF_EQUAL_DEC);
                if(info.NWhispers(day) > 0){
                    if(day >= DAY_EXECUTE_FIRST){
                        if(d1 == info.whisperVoteDec[day][player]){ // 同じ
                            FooC(FEA_VOTE_WEREWOLF_EQUAL_DEC, base);
                            FooXC(FEA_VOTE_WEREWOLF_EQUAL_DEC, base + 1,
                                  rate(info.whisperVoteEqualDecNum, info.whisperVoteDecDayNum));
                        }
                    }
                }
            }*/
            
            // 宣言通りの相手を襲撃
            /*if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_ATTACK_VOTE_EQUAL_DEC);
                if(day >= DAY_EXECUTE_FIRST && day >= DAY_ATTACK_FIRST){
                    if(d1 == info.attackVoteDec[day][player]){ // 同じ
                        FooC(FEA_ATTACK_VOTE_EQUAL_DEC, base);
                        FooXC(FEA_ATTACK_VOTE_EQUAL_DEC, base + 1,
                              rate(info.attackVoteEqualDecNum, info.attackVoteDecDayNum));
                    }
                }
            }*/
            
            // 前日と同じ相手に襲撃投票する
            /*if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_ATTACK_VOTE_EQUAL_YESTERDAY);
                if(day >= DAY_ATTACK_FIRST + 1){
                    if(d1 == info.attackVote(day - 1, player)){
                        FooC(FEA_ATTACK_VOTE_EQUAL_YESTERDAY, base);
                    }
                }
            }*/
            // 占い師CO者に襲撃投票する
            if(R == Role::WEREWOLF){
                constexpr int base = FEA_IDX(FEA_ATTACK_VOTE_SEER);
                if(day >= DAY_ATTACK_FIRST){
                    if(info.comingOutRole[d1] == Role::SEER){ // 占い師CO者
                        int coNum = info.comingOutedNum[Role::SEER];
                        if(coNum > 1){ // 複数CO
                            FooC(FEA_ATTACK_VOTE_SEER, base + 1);
                        }else{
                            FooC(FEA_ATTACK_VOTE_SEER, base);
                        }
                    }
                }
            }
            
            double exps = exp(s / pol.temperature());
            FASSERT(exps, cerr << "s = " << s << " T = " << pol.temperature() << endl;);
            pol.template feedCandidateScore<M>(exps);
            
            dst[m] = exps;
            sum += exps;
        }
        return sum;
	}
	
#undef FooXC
#undef FooC
#undef FooX
#undef Foo
    
    template<int M = 1, class move_t, class info_t, class policy_t>
    double calcActionPolicyScoreWithRole(double *const dst, const move_t *const mv, const int NMoves,
                                       const info_t& info, const int player, const int role, const policy_t& pol){
        switch(role){
            case 0: return calcActionPolicyScore<0, M>(dst, mv, NMoves, info, player, pol); break;
            case 1: return calcActionPolicyScore<1, M>(dst, mv, NMoves, info, player, pol); break;
            case 2: return calcActionPolicyScore<2, M>(dst, mv, NMoves, info, player, pol); break;
            case 3: return calcActionPolicyScore<3, M>(dst, mv, NMoves, info, player, pol); break;
            case 4: return calcActionPolicyScore<4, M>(dst, mv, NMoves, info, player, pol); break;
            case 5: return calcActionPolicyScore<5, M>(dst, mv, NMoves, info, player, pol); break;
            default: UNREACHABLE; break;
        }
        return 0;
    }
}

#endif // AIWOLF_THOUSAND_ACTIONPOLICY_HPP_