/*
 biEstimation.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_BIESTIMATION_HPP_
#define AIWOLF_THOUSAND_BIESTIMATION_HPP_

// 人 x 人の関係で、どの役職の組み合わせかを推定

namespace AIWolf{
    namespace Thousand{
        /**************************逆方向評価点**************************/
        
        namespace BiEstimationSpace{
            enum{
                FEA_VOTE = 0, // 投票した
                FEA_DIVINE, // 占った
                FEA_INQUEST, // 霊媒した
                FEA_ESTIMATE, // 推定した
                FEA_
                
                FEA_ALL,
            };
            
            constexpr int numTable[FEA_ALL] = {
                ipow(N_ROLES, 2),
                ipow(N_ROLES, 3),
                ipow(N_ROLES, 3),
                ipow(N_ROLES, 3),
            };
            constexpr int FEA_NUM(unsigned int fea){
                return numTable[fea];
            }
            constexpr int FEA_IDX(unsigned int fea){
                return (fea == 0) ? 0 : (FEA_IDX(fea - 1) + FEA_NUM(fea - 1));
            }
            constexpr int FEA_NUM_ALL = FEA_IDX(FEA_ALL);
            
#define LINEOUT(fea, str) { out << str << endl; int base = FEA_IDX(fea); for (int i = 0; i < FEA_NUM(fea); ++i){os(base + i); } out << endl; }
            
#define LINEOUTX(fea, str, x) { out << str << endl; int base = FEA_IDX(fea); int num = FEA_NUM(fea);for (int i = 0;;){os(base + i); ++i; if(i >= num)break; if(i % (x) == 0){ out << endl; }} out << endl; }
            
            int commentParam(std::ostream& out, const double param[FEA_NUM_ALL]){
                auto os = [&out, param](int idx)->void{ out << param[idx] << " "; };
                
                out << "****** BI ESTIMATION ******" << endl;
                
                //LINEOUT(,);
                return 0;
            }
#undef LINEOUT
        }
        
        using BiEstimator = SoftmaxPolicy<BiEstimationSpace::FEA_NUM_ALL, 1>;
        using BiEstimatorLearner = SoftmaxPolicyLearner<BiEstimator>;
        
        int foutComment(const Evaluator& est, const std::string& fName){
            std::ofstream ofs(fName, std::ios::out);
            return BiEstimationSpace::commentParam(ofs, est.param_);
        }
        
#define Foo(i) {s += est.param(i); est.template feedFeatureScore<M>((i), 1.0);}
#define FooX(i, x) s += est.param(i) * (x); est.template feedFeatureScore<M>((i), (x));}
        
#define FooC(p, i) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
        cerr << "FooC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
        cerr << " (pi = " << (p) << ")" << endl; );\
        s += est.param(i); est.template feedFeatureScore<M>((i), 1.0);}
        
#define FooXC(p, i, x) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
        cerr << "FooXC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
        cerr << " (pi = " << (p) << ")" << endl; );\
        s += est.param(i) * (x); est.template feedFeatureScore<M>((i), (x));}
        
        
        template<int R, int M = 1, class info_t, class estimator_t>
        void calcBiEstimationScore(double *const score, const info_t& info, int p0, int p1, const estimator_t& est){
            
            using namespace BiEstimationSpace;
            
            est.template initCalculatingScore<M>(N_ROLES * N_ROLES);
            
            const int day = info.day();
            
            // 全ての役職に対して検討
            for (int r0 = 0; r0 < N_ROLES; ++r0){
                for(int r1 = 0; r1 < N_ROLES; ++r1){
                    
                    int ir = r0 * N_ROLES + r1;
                    
                    eval.template initCalculatingCandidateScore<M>();
                    double s = 0;
                    
                    // 投票
                    {
                        constexpr int base = FEA_IDX(FEA_VOTE);
                        for(int d = DAY_EXECUTE_FIRST; d < day; ++d){
                            if(info.vote(d, p0) == p1){
                                FooC(FEA_VOTE, base + N_ROLES * r0 + r1);
                            }
                        }
                    }
                    
                    // 占い
                    {
                        constexpr int base = FEA_IDX(FEA_DIVINE);
                        for(int d = DAY_DIVINE_FIRST; d < day; ++d){
                            if(info.divineResult[p0][p1] == p1){
                                FooC(FEA_DIVINE, base + N_ROLES * r0 + r1);
                            }
                        }
                    }
                    
                    // 霊媒
                    {
                        constexpr int base = FEA_IDX(FEA_INQUEST);
                        for(int d = DAY_ATTACK_FIRST; d < day; ++d){
                            if(info.vote(d, p0) == p1){
                                FooC(FEA_INQUEST, base + N_ROLES * r0 + r1);
                            }
                        }
                    }
                }
                
                double exps = exp(s / est.temperature());
                FASSERT(exps, cerr << "s = " << s << " T = " << eval.temperature() << endl;);
                eval.template feedCandidateScore<M>(exps);
                
                dst[r] = exps;
            }
            return 0;
        }
        
#endif // AIWOLF_THOUSAND_SINGLEESTIMATION_HPP_