/*
 evaluation.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_EVALUATION_HPP_
#define AIWOLF_THOUSAND_EVALUATION_HPP_

#include "../aiwolf.hpp"
#include "./thinkInfo.hpp"

namespace AIWolf{

    /**************************評価点**************************/
    
    // その日の開始時点での情報から計算
    
    namespace EvaluationSpace{
        enum{
            FEA_TEAM_PLAYERS = 0, // チーム人数
            FEA_SPECIES_PLAYERS, // 種族人数
            FEA_ROLE_PLAYERS,
            
            FEA_SEER_ALIVE,
            FEA_MEDIUM_ALIVE,
            FEA_BODYGUARD_ALIVE,
            
            FEA_CO_POSSESSED_ALIVE, // COしている狂人
            
            FEA_ALL,
        };
        
        constexpr int numTable[] = {
            (N_MAX_WEREWOLF_TEAM + 1) * (N_MAX_VILLAGER_TEAM + 1),
            (N_MAX_WEREWOLF_SPECIES + 1) * (N_MAX_HUMAN_SPECIES + 1),
            0, //mul(BASE_RULE.rolePlayers),
            N_MAX_DAYS,
            N_MAX_DAYS,
            N_MAX_DAYS,
            N_MAX_DAYS * 2,
        };
        static_assert(FEA_ALL == sizeof(numTable) / sizeof(int), "num of features confliction.");
        
        constexpr int FEA_NUM(unsigned int fea){
            return numTable[fea];
        }
        constexpr int FEA_IDX(unsigned int fea){
            return (fea == 0) ? 0 : (FEA_IDX(fea - 1) + FEA_NUM(fea - 1));
        }
        constexpr int FEA_NUM_ALL = FEA_IDX(FEA_ALL);
 
#define LINEOUT(fea, str) { out << str << endl; int base = FEA_IDX(fea); for (int i = 0; i < FEA_NUM(fea); ++i){os(base + i); } out << endl; }
        
#define LINEOUTX(fea, str, x) { out << str << endl; int base = FEA_IDX(fea); int num = FEA_NUM(fea);\
for (int i = 0;;){os(base + i); ++i; if(i >= num)break; if(i % (x) == 0){ out << endl; }} out << endl; }
        
        int commentParam(std::ostream& out, const double param[FEA_NUM_ALL]){
            auto os = [&out, param](int idx)->void{ out << param[idx] << " "; };

            out << "****** EVALUATION ******" << endl;
            
            LINEOUTX(FEA_TEAM_PLAYERS, "NTeamPlayers", N_MAX_VILLAGER_TEAM + 1);
            LINEOUTX(FEA_SPECIES_PLAYERS, "NTeamPlayers", N_MAX_HUMAN_SPECIES + 1);
            
            LINEOUTX(FEA_SEER_ALIVE, "Seer Alive", N_MAX_DAYS);
            LINEOUTX(FEA_MEDIUM_ALIVE, "Medium Alive", N_MAX_DAYS);
            LINEOUTX(FEA_BODYGUARD_ALIVE, "BodyGuard Alive", N_MAX_DAYS);
            
            LINEOUTX(FEA_CO_POSSESSED_ALIVE, "CO - Possessed alive", N_MAX_DAYS);
            
            return 0;
        }
#undef LINEOUT
    }
    
    using Evaluator = SoftmaxPolicy<EvaluationSpace::FEA_NUM_ALL, 1>;
    using EvaluatorLearner = SoftmaxPolicyLearner<Evaluator>;
    
    int foutComment(const Evaluator& ev, const std::string& fName){
        std::ofstream ofs(fName, std::ios::out);
        return EvaluationSpace::commentParam(ofs, ev.param_);
    }
    
#define Foo(i) {s += eval.param(i); eval.template feedFeatureScore<M>((i), 1.0);}
#define FooX(i, x) s += eval.param(i) * (x); eval.template feedFeatureScore<M>((i), (x));}
    
#define FooC(p, i) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
    cerr << "FooC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
    cerr << " (pi = " << (p) << ")" << endl; );\
    s += eval.param(i); ev.template feedFeatureScore<M>((i), 1.0);}
    
#define FooXC(p, i, x) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
    cerr << "FooXC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
    cerr << " (pi = " << (p) << ")" << endl; );\
    s += eval.param(i) * (x); eval.template feedFeatureScore<M>((i), (x));}

    template<int M = 1, class info_t, class evaluator_t>
    void calcEvaluationScore(double *const dst, const info_t& info, const evaluator_t& eval){
    
        using namespace EvaluationSpace;
        
        eval.template initCalculatingScore<M>(2);
        
        const int day = info.day();
        
        for(int w = 0; w < N_TEAMS; ++w){
            
            eval.template initCalculatingCandidateScore<M>();
            
            const int wl = 1 - 2 * w; // 村人陣営勝ちなら +1, 人狼陣営勝ちなら -1
            
            double s = 0;
            
            // 各陣営の残り人数
            {
                constexpr int base = FEA_IDX(FEA_TEAM_PLAYERS);
                int i = base + info.NTeamAlive(day, Team::WEREWOLF) * (N_MAX_VILLAGER_TEAM + 1) + info.NTeamAlive(day, Team::VILLAGER);
                FooXC(FEA_TEAM_PLAYERS, i, wl);
            }
            // 各種族の残り人数
            {
                constexpr int base = FEA_IDX(FEA_SPECIES_PLAYERS);
                int i = base + info.NSpeciesAlive(day, Species::WEREWOLF) * (N_MAX_HUMAN_SPECIES + 1) + info.NTeamAlive(day, Species::HUMAN);
                FooXC(FEA_SPECIES_PLAYERS, i, wl);
            }
            
            // 各役職の残り人数
            /*{
                constexpr int base = EV_IDX(FEA_ROLE_PLAYERS);
                int i = (info.NTeamAlivs(Team::VILLAGER) - 1)
                * (info.NTeamAlivs(Team::WEREWOLF) - 1);
                FooC(EV_SCORE, i);
            }*/
            
            // 占い師生き
            {
                constexpr int base = FEA_IDX(FEA_SEER_ALIVE);
                if(info.roleAlive(day, Role::SEER)){
                    FooXC(FEA_SEER_ALIVE, base + day, wl);
                }
            }
            // 霊媒師生き
            {
                constexpr int base = FEA_IDX(FEA_MEDIUM_ALIVE);
                if(info.roleAlive(day, Role::MEDIUM)){
                    FooXC(FEA_MEDIUM_ALIVE, base + day, wl);
                }
            }
            // 狩人生き
            {
                constexpr int base = FEA_IDX(FEA_BODYGUARD_ALIVE);
                if(info.roleAlive(day, Role::BODYGUARD)){
                    FooXC(FEA_BODYGUARD_ALIVE, base + day, wl);
                }
            }
            // CO狂人生き
            {
                constexpr int base = FEA_IDX(FEA_CO_POSSESSED_ALIVE);
                if(info.roleAlive(day, Role::POSSESSED)){
                    int p = info.rolePlayers(Role::POSSESSED).bsf();
                    if(info.comingOutRole[p] == Role::SEER){
                        FooXC(FEA_CO_POSSESSED_ALIVE, base + day, wl);
                    }else if(info.comingOutRole[p] == Role::MEDIUM){
                        FooXC(FEA_CO_POSSESSED_ALIVE, base + N_MAX_DAYS + day, wl);
                    }
                }
            }
            
            
            double exps = exp(s / eval.temperature());
            FASSERT(exps, cerr << "s = " << s << " T = " << eval.temperature() << endl;);
            eval.template feedCandidateScore<M>(exps);
            
            dst[w] = exps;
        }
	}
	
#undef FooXC
#undef FooC
#undef FooX
#undef Foo
}

#endif // AIWOLF_THOUSAND_EVALUATION_HPP_