/*
 playout.hpp
 Katsuki Ohto
*/

#ifndef AIWOLF_THOUSAND_PLAYOUT_HPP_
#define AIWOLF_THOUSAND_PLAYOUT_HPP_

namespace AIWolf{
    namespace Thousand{
        
        template<class info_t, class estimator_t, class evaluator_t, class policy_t, class dice_t>
        double doSimulation(info_t& info, BitArray64<4, 16> sample, int player, BitArray32<16, 2> act,
                         const estimator_t& est,
                         const evaluator_t& eval,
                         const policy_t& pol,
                         dice_t *const pdice){
            
            const int day = info.day();
            
            const int P = info.rule().players;
            BitArray32<16, 2> mv[N_MAX_PLAYERS * N_MAX_PLAYERS]; // 行動生成バッファ
            double score[N_MAX_PLAYERS * N_MAX_PLAYERS]; // 行動方策得点
            
            // 推定情報に従ってプレーヤーを配置する
            info.setSampleRoles(sample);
            
            int win = info.isGameOver(info.day());
            if(win != NONE_TEAM){
                //cerr << "dealt mate " << day << sample << endl;
                return -9999; } // 配置の時点で詰んでいたら無効

            // 各プレーヤーの行動決定
            BitArray64<4, 16> votedNum(0); // 処刑投票結果
            BitArray64<4, 16> attackVotedNum(0); // 処刑投票結果
            int execute = -1;
            int divine = -1; // 占い
            int guard = -1; // 護衛
            int attack = -1;
            int seerNum, bodyGuardNum;
            std::pair<int, int> attackVote[maxPlayerNum(Role::WEREWOLF)];
            int attackVoteCnt = 0;
            
            for(int p = 0; p < P; ++p){
                if(info.isAlive(day, p)){
                    BitArray32<16, 2> tmp;
                    if(p == player){ // シミュレーションオーナーの行動
                        tmp = act;
                    }else{
                        const int NMoves = genValidActions(mv, info, p); // 行動生成
                        if(NMoves == 0){ continue; }
                        // 各プレーヤーの視点で役職推定
                        //estimateNow(&info.sub[p], p, info, est);
                        // 行動に得点をつける
                        double sum = calcActionPolicyScoreWithRole<0>(score, mv, NMoves, info, p, info.playerRole(p), pol);
                        // 温度付きで選択
                        double dr = pdice->drand() * sum;
                        int chosen = NMoves - 1;
                        for(int m = 0; m < NMoves; ++m){
                            //cerr << mv[m] << score[m] << endl;
                            dr -= score[m];
                            if(dr <= 0){ chosen = m; break; }
                        }
                        tmp = mv[chosen];
                        //cerr << "chosen = " << tmp << endl;
                    }
                    if(day >= DAY_EXECUTE_FIRST){
                        votedNum.add(tmp[0], 1); // 投票処理
                    }
                    
                    switch(info.playerRole(p)){
                        case Role::SEER: if(day >= DAY_DIVINE_FIRST){
                            divine = tmp[1]; seerNum = p;
                            ASSERT(info.isValidDivine(day, p, divine), cerr << divine << endl;);
                        } break;
                        case Role::BODYGUARD: if(day >= DAY_ATTACK_FIRST){
                            guard = tmp[1]; bodyGuardNum = p;
                            ASSERT(info.isValidGuard(day, p, guard), cerr << guard << endl;);
                        } break;
                        case Role::WEREWOLF: if(day >= DAY_ATTACK_FIRST){
                            attackVote[attackVoteCnt] = std::make_pair(p, tmp[1]);
                            ++attackVoteCnt;
                            ASSERT(info.isValidAttackVote(day, p, tmp[1]), cerr << tmp[1] << endl;);
                        } break;
                        default: break;
                    }
                }
            }
            
            // 各プレーヤーの行動が出揃った
            { // 処刑決定
                int best = -1;
                int bestCount = 2;
                int bestNum = 0;
                
                for(int p = 0; p < P; ++p){
                    if(votedNum[p] > bestNum){
                        best = p; bestNum = votedNum[p]; bestCount = 2;
                    }else if(votedNum[p] == bestNum){
                        if(pdice->rand() % bestCount == 0){
                            best = p;
                        }
                        ++bestCount;
                    }
                }
                execute = best;
                CERR << "votedNum = " << votedNum << endl;
                CERR << "execute = " << execute << endl;
                ASSERT(info.isAlive(day, execute), cerr << execute << endl;);
            }
            { // 襲撃決定
                int best = -1;
                int bestCount = 2;
                int bestNum = 0;
                
                int realAttackVoteCnt = 0;
                for(int n = 0; n < attackVoteCnt; ++n){
                    int p = attackVote[n].first;
                    int t = attackVote[n].second;
                    if(execute == p){ continue; } // 先に処刑された場合には襲撃投票に参加できない
                    attackVotedNum.add(t, 1);
                    ++realAttackVoteCnt;
                }
                for(int p = 0; p < P; ++p){
                    if(attackVotedNum[p] > bestNum){
                        best = p; bestNum = attackVotedNum[p]; bestCount = 2;
                    }else if(attackVotedNum[p] == bestNum){
                        if(pdice->rand() % bestCount == 0){
                            best = p;
                        }
                        ++bestCount;
                    }
                }
                attack = best;
                CERR << "attack = " << attack << endl;
                if(realAttackVoteCnt > 0){
                    ASSERT(info.isValidAttack(day, attack), cerr << attack << endl;);
                }
            }
            if(attack != -1){
                info.setAttack(day, attack);
                if(guard != -1){ // 護衛
                    if(guard != execute){ // 狩人が処刑されていない
                        info.setGuard(day, bodyGuardNum, guard);
                        if(guard == attack){ // 護衛に遭った
                            attack = -1; // 襲撃無効化
                        }
                    }else{
                        guard = -1; // 護衛不可
                    }
                }
            }
            if(divine != -1){
                if(seerNum == execute || seerNum == attack){
                    // 占い師死亡により占い不可
                    divine = -1;
                }
            }
            
            // 試合進行処理
            if(execute != -1){
                info.setExecuted(day, execute);
                if(info.roleAlive(day, Role::MEDIUM)){ // 霊媒師が生きている
                    info.setInquested(day, info.roleAlive(day, Role::MEDIUM).bsf(), execute,
                                      getSpecies(info.playerRole(execute)));
                }
            }
            if(attack != -1){ // 襲撃成功
                info.setAttacked(day, attack);
            }
            if(divine != -1){ // 占い成功
                info.setDivine(day, seerNum, divine);
                info.setDivined(day, seerNum, divine, getSpecies(info.playerRole(divine)));
            }
            
            // 試合進行が終わったので、次の日の朝の状態を評価する
            info.proceedDay(day + 1); // 状態更新
            win = info.isGameOver(day + 1);
            if(win != NONE_TEAM){ // 試合終了した
                // 村人勝ちなら+1、人狼勝ちなら-1を返す
                if(win == Team::VILLAGER){
                    return +1;
                }else{
                    return -1;
                }
            }else{ // 試合続行のとき
                // 評価関数を呼ぶ
                double prob[2];
                calcEvaluationScore<0>(prob, info, eval); // 評価計算
                double wr = prob[Team::VILLAGER] / (prob[0] + prob[1]); // 村人側の勝率
                return wr * 2 - 1;
            }
        }
    }
}

#endif // AIWOLF_THOUSAND_PLAYOUT_HPP_
