/*
 singleEstimation.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_SINGLEESTIMATION_HPP_
#define AIWOLF_THOUSAND_SINGLEESTIMATION_HPP_

#include "../aiwolf.hpp"
#include "./thinkInfo.hpp"

// 一人の人について、どの役職の可能性が高そうか推定

namespace AIWolf{
	namespace Thousand{
        /**************************逆方向評価点**************************/
        
        namespace EstimationSpace{
            enum{
                // 共通情報からの推論
                FEA_COMINGOUT = 0, // カミングアウト
                FEA_ATTACK_FAILURE, // 襲撃が起こらなかった
                FEA_DIVINED_RESULT, // 占われた
                FEA_INQUESTED_RESULT, // 霊媒された
                FEA_SEER_DEATH, // 占い師CO者の死に方
                FEA_MEDIUM_DEATH, // 霊媒師CO者の死に方
                FEA_DIVINE_ATTACKED, // 人狼と占った相手が襲撃された
                FEA_TALK_NUM, // 有効会話回数
                FEA_VOTE_EQUAL_DEC, // 投票宣言通りに投票したか
                
                // 主観情報の利用
                FEA_COMINGOUTED_MY_ROLE, // 自分の役職を騙る人(自分は1人の役職)
                FEA_COMINGOUTED_WEREWOLF, // 自分が人狼のときに、人狼を騙る人
                FEA_DIVINED_CONFLICT_MINE_HUMAN, // 占い師が自分(人間)を占ったが、占い結果が自分の種族と矛盾
                FEA_DIVINED_CONFLICT_WEREWOLF, // 自分が人狼のときに、占い結果が違う
                FEA_INQUESTED_CONFLICT_WEREWOLF, // 自分が人狼のときに、霊媒結果が違う
                FEA_DIVINED_CONFLICT_MY_INQUEST, // 占い師が、自分の霊媒結果と矛盾することを報告した
                FEA_INQUESTED_CONFLICT_MY_DIVINE, // 霊媒師が、自分の占い結果と矛盾することを報告した
                
                FEA_ATTACK_FAILURE_WEREWOLF, // 狼視点の襲撃失敗
                
                FEA_VOTE_WEREWOLF, // 人狼に投票したか
                
                FEA_ALL,
            };
            
            constexpr int numTable[] = {
                N_ROLES * N_ROLES + 2,
                1,
                N_ROLES * 2,
                N_ROLES * 2,
                N_ROLES * 2,
                N_ROLES * 2,
                2,
                N_ROLES,
                N_ROLES,
                
                4 * N_ROLES,
                N_ROLES,
                N_ROLES,
                2,
                2,
                N_ROLES,
                N_ROLES,
                
                3,
                
                3,
            };
            static_assert(FEA_ALL == sizeof(numTable) / sizeof(int), "num of features confliction.");
            
            constexpr int FEA_NUM(unsigned int fea){
                return numTable[fea];
            }
            constexpr int FEA_IDX(unsigned int fea){
                return (fea == 0) ? 0 : (FEA_IDX(fea - 1) + FEA_NUM(fea - 1));
            }
            constexpr int FEA_NUM_ALL = FEA_IDX(FEA_ALL);
            
#define LINEOUT(fea, str) { out << str << endl; int base = FEA_IDX(fea); for (int i = 0; i < FEA_NUM(fea); ++i){os(base + i); } out << endl; }
            
#define LINEOUTX(fea, str, x) { out << str << endl; int base = FEA_IDX(fea); int num = FEA_NUM(fea);for (int i = 0;;){os(base + i); ++i; if(i >= num)break; if(i % (x) == 0){ out << endl; }} out << endl; }
            
            int commentParam(std::ostream& out, const double param[FEA_NUM_ALL]){
                auto os = [&out, param](int idx)->void{ out << param[idx] << " "; };
                
                out << "****** ESTIMATION ******" << endl;
                
                LINEOUTX(FEA_COMINGOUT, "ComingOut", N_ROLES);
                LINEOUT(FEA_ATTACK_FAILURE, "AttackFailure");
                LINEOUTX(FEA_DIVINED_RESULT, "DivinedResult", N_ROLES);
                LINEOUTX(FEA_INQUESTED_RESULT, "InquestedResult", N_ROLES);
                LINEOUTX(FEA_SEER_DEATH, "SeerDeath", N_ROLES);
                LINEOUTX(FEA_MEDIUM_DEATH, "MediumDeath", N_ROLES);
                LINEOUT(FEA_DIVINE_ATTACKED, "DivineAttacked");
                LINEOUTX(FEA_TALK_NUM, "TalkNum", N_ROLES);
                LINEOUTX(FEA_VOTE_EQUAL_DEC, "VoteEqualDec", N_ROLES);
                
                LINEOUTX(FEA_COMINGOUTED_MY_ROLE, "ComingOutConflictAgainstMyRole", N_ROLES);
                LINEOUTX(FEA_COMINGOUTED_WEREWOLF, "ComingOutConflictWerewolf", N_ROLES);
                LINEOUTX(FEA_DIVINED_CONFLICT_MINE_HUMAN, "DivineConflictMineHuman", N_ROLES);
                LINEOUT(FEA_DIVINED_CONFLICT_WEREWOLF, "DivineConflictWerewolf");
                LINEOUT(FEA_INQUESTED_CONFLICT_WEREWOLF, "InquestConflictWerewolf");
                LINEOUTX(FEA_DIVINED_CONFLICT_MY_INQUEST, "DivineConflictMyInquest", N_ROLES);
                LINEOUTX(FEA_INQUESTED_CONFLICT_MY_DIVINE, "InquestConflictMyDivine", N_ROLES);
                //LINEOUT(,);
                return 0;
            }
#undef LINEOUT
        }
        
        using Estimator = SoftmaxPolicy<EstimationSpace::FEA_NUM_ALL, N_ROLES + 1>;
        using EstimatorLearner = SoftmaxPolicyLearner<Estimator>;
        
        int foutComment(const Estimator& est, const std::string& fName){
            std::ofstream ofs(fName, std::ios::out);
            return EstimationSpace::commentParam(ofs, est.param_);
        }
        
#define Foo(i) {s += est.param(i); est.template feedFeatureScore<M>((i), 1.0);}
#define FooX(i, x) s += est.param(i) * (x); est.template feedFeatureScore<M>((i), (x));}
        
#define FooC(p, i) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
        cerr << "FooC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
        cerr << " (pi = " << (p) << ")" << endl; );\
        s += est.param(i); est.template feedFeatureScore<M>((i), 1.0);}
        
#define FooXC(p, i, x) {ASSERT(FEA_IDX(p) <= i && i < FEA_IDX((p) + 1),\
        cerr << "FooXC() : illegal index " << (i) << " in " << FEA_IDX(p) << "~" << (FEA_IDX((p) + 1) - 1);\
        cerr << " (pi = " << (p) << ")" << endl; );\
        s += est.param(i) * (x); est.template feedFeatureScore<M>((i), (x));}
        
        
        template<int R, int M = 1, class info_t, class estimator_t>
        void calcEstimationScore(double *const dst, const info_t& info, int myId, int player, const estimator_t& est){
            
            // 人同士の関係を見ない単ナイーブな役職確率分布を計算
            
            using namespace EstimationSpace;
            
            est.template initCalculatingScore<M>(N_ROLES);
            
            const int day = info.day();
            const int P = info.rule().players;
            
            //cerr << day << "," << P << endl;
            
            // 全ての役職に対して検討
            for (int r = 0; r < N_ROLES; ++r){
                est.template initCalculatingCandidateScore<M>();
                double s = 0;
                
                // 襲撃を受けた場合に人狼でない(学習対象外)
                if(info.causeOfDeath(player) == ATTACKED && r == Role::WEREWOLF){
                    //cerr << "at-nw" << endl; getchar();
                    s = -DBL_MAX; goto END_CANDIDATE;
                }
                
                // 定数項(学習対象外)
                s += info.rule().rolePlayers[r];
                
                // カミングアウト情報
                {
                    constexpr int base = FEA_IDX(FEA_COMINGOUT);
                    if(info.anyComingOut(player)){ // COしていた場合
                        int cr = info.comingOutRole[player];
                        // 役職 x 役職に応じてベース点
                        FooC(FEA_COMINGOUT, base + cr * N_ROLES + r);
                        
                        if(r == Role::SEER){ // 占い師
                            if(info.comingOutedNum[r] > 1){ // 複数人がCOしている
                                FooC(FEA_COMINGOUT, base + N_ROLES * N_ROLES);
                            }
                            // 何番目にCOしたか
                        }else if(r == Role::MEDIUM){
                            if(info.comingOutedNum[r] > 1){ // 複数人がCOしている
                                FooC(FEA_COMINGOUT, base + N_ROLES * N_ROLES + 1);
                            }
                        }
                    }
                }
                
                // 襲撃成功or失敗情報
                {
                    constexpr int base = FEA_IDX(FEA_ATTACK_FAILURE);
                    if(R != Role::WEREWOLF && r == Role::BODYGUARD){
                        for(int d = DAY_ATTACK_FIRST; d < day; ++d){
                            if(!info.anyAttacked(d)){ // 襲撃が起こらなかった
                                if(info.isAliveAfterExecution(d, player)){
                                    // 襲撃時点で生きていた場合に狩人であるか
                                    FooC(FEA_ATTACK_FAILURE, base);
                                }
                            }
                        }
                    }
                }
                
                // 占い師の占い結果報告
                if(R != Role::SEER){
                    constexpr int base = FEA_IDX(FEA_DIVINED_RESULT);
                    int coNum = info.comingOutedNum[Role::SEER];
                    if(examValidRole(R) && info.comingOutRole[myId] == Role::SEER){
                        --coNum; // 自分を抜く
                    }
                    if(coNum > 0){ // 占い師COが複数
                        int idx = base + ((coNum > 1) ? N_ROLES : 0); // CO者が複数か
                        int dn[2] = { (int)info.divinedNum[0][player], (int)info.divinedNum[1][player] };
                        
                        // 占い結果に自分の報告が含まれる場合は抜く
                        // 自分自身に対する報告はそもそもカウントしていない
                        if(examValidRole(R) && myId != player && info.divineResult[myId][player] != NONE_SPECIES){
                            --dn[info.divineResult[myId][player]];
                        }
                        ASSERT(dn[0] >= 0 && dn[1] >= 0,
                               cerr << dn[0] << "," << dn[1] << endl;
                               cerr << myId << " " << R << endl;
                               cerr << info.toSurveyString(day););
                        if(dn[0] || dn [1]){ // 占われている
                            double rate = double(dn[0] - dn[1])
                            / (dn[0] + dn[1]); // -1 ~ +1
                            /*cerr << rate << endl;
                            if(rate > 1){
                                cerr << player << endl;
                                cerr << info.divinedNum[0] << info.divinedNum[1] << endl;
                                cerr << (info.divinedNum[0][player] - info.divinedNum[1][player]) << endl;
                                getchar();
                            }*/
                            ASSERT(-1 <= rate && rate <= +1, cerr << rate << endl;);
                            FooXC(FEA_DIVINED_RESULT, idx + r, rate);
                        }
                    }
                }
                
                // 霊媒師の霊媒結果報告(霊媒された側)
                if(R != Role::MEDIUM){
                    constexpr int base = FEA_IDX(FEA_INQUESTED_RESULT);
                    
                    int coNum = info.comingOutedNum[Role::MEDIUM];
                    if(examValidRole(R) && info.comingOutRole[myId] == Role::MEDIUM){
                        --coNum; // 自分を抜く
                    }
                    
                    if(coNum > 0){ // 霊媒師COがいる
                        int idx = base + ((coNum > 1) ? N_ROLES : 0); // CO者が複数か
                        
                        int in[2] = { (int)info.inquestedNum[0][player], (int)info.inquestedNum[1][player] };
                        
                        // 霊媒結果に自分の報告が含まれる場合は抜く
                        // 自分自身に対する報告はそもそもカウントしていない
                        if(examValidRole(R) && myId != player && info.inquestResult[myId][player] != NONE_SPECIES){
                            --in[info.inquestResult[myId][player]];
                        }
                        ASSERT(in[0] >= 0 && in[1] >= 0,
                               cerr << in[0] << "," << in[1] << endl;
                               cerr << myId << " " << R << endl;
                               cerr << info.toSurveyString(day););
                        if(in[0] || in[1]){ // 霊媒されている
                            double rate = double(in[0] - in[1])
                            / (in[0] + in[1]); // -1 ~ +1
                            
                            ASSERT(-1 <= rate && rate <= +1, cerr << rate << endl;);
                            FooXC(FEA_INQUESTED_RESULT, idx + r, rate);
                        }
                    }
                }
                
                // 占い師CO者の死に方
                if(R != Role::SEER){
                    constexpr int base = FEA_IDX(FEA_SEER_DEATH);
                    if(info.comingOutRole[player] == Role::SEER){
                        if(info.causeOfDeath(player) == ATTACKED){
                            FooC(FEA_SEER_DEATH, base + r);
                        }else if(info.causeOfDeath(player) == EXECUTED){
                            FooC(FEA_SEER_DEATH, base + N_ROLES + r);
                        }
                    }
                }
                
                // 霊媒CO者の死に方
                if(R != Role::MEDIUM){
                    constexpr int base = FEA_IDX(FEA_MEDIUM_DEATH);
                    if(info.comingOutRole[player] == Role::MEDIUM){
                        if(info.causeOfDeath(player) == ATTACKED){
                            FooC(FEA_MEDIUM_DEATH, base + r);
                        }else if(info.causeOfDeath(player) == EXECUTED){
                            FooC(FEA_MEDIUM_DEATH, base + N_ROLES + r);
                        }
                    }
                }
                
                // 人狼と占った相手が襲撃されたとき、自分が占い師か、狂人かどうか
                if(R != Role::SEER){
                    constexpr int base = FEA_IDX(FEA_DIVINE_ATTACKED);
                    if(info.comingOutRole[player] == Role::SEER){
                        int f = 0;
                        for(int t = 0; t < P; ++t){
                            if(info.divineResult[player][t] == Species::WEREWOLF
                               && info.causeOfDeath(t) == ATTACKED){
                                f = 1; break;
                            }
                        }
                        if(f){
                            if(r == Role::SEER){
                                FooC(FEA_DIVINE_ATTACKED, base);
                            }else if(r == Role::POSSESSED){
                                FooC(FEA_DIVINE_ATTACKED, base + 1);
                            }
                        }
                    }
                }
                
                // 有効会話回数
                {
                    constexpr int base = FEA_IDX(FEA_TALK_NUM);
                    int days = (info.deathDay(player) == 0) ? (day + 1) : (info.deathDay(player) + 1);
                    double n = info.anyTalkNum[player].sum() / double(days);
                    FooXC(FEA_TALK_NUM, base + r, sqrt(n));
                }
                
                // 表の投票宣言通りに投票した回数
                {
                    constexpr int base = FEA_IDX(FEA_VOTE_EQUAL_DEC);
                    double prob = !info.voteDecDayNum[player] ? 0.5 :
                    info.voteEqualDecNum[player] / info.voteDecDayNum[player];
                    FooXC(FEA_VOTE_EQUAL_DEC, base + r, prob);
                }
                
                // <主観情報>
                if(examValidRole(R)){
                    // 自分の役職を騙る人(自分は1人の役職)
                    if(maxPlayerNum(R) == 1){
                        constexpr int base = FEA_IDX(FEA_COMINGOUTED_MY_ROLE);
                        
                        if(info.comingOutRole[player] == R){
                            int idx = base + (R - 1) * N_ROLES + r;
                            FooC(FEA_COMINGOUTED_MY_ROLE, idx);
                        }
                    }
                    
                    // 自分の役職を騙る人(自分は人狼)
                    if(R == Role::WEREWOLF){
                        constexpr int base = FEA_IDX(FEA_COMINGOUTED_WEREWOLF);
                        if(info.comingOutRole[player] == Role::WEREWOLF){
                            FooC(FEA_COMINGOUTED_WEREWOLF, base + r);
                        }
                    }
                    
                    // 占い師が自分(人間)を占ったが、占い結果が自分の種族と矛盾
                    if(getSpecies(R) == Species::HUMAN){
                        constexpr int base = FEA_IDX(FEA_DIVINED_CONFLICT_MINE_HUMAN);
                        
                        if(info.divineResult[player][myId] == Species::WEREWOLF){
                            FooC(FEA_DIVINED_CONFLICT_MINE_HUMAN, base + r);
                        }
                    }
                    
                    // 自分が人狼のときに、占い結果が違う
                    if(R == Role::WEREWOLF){
                        constexpr int base = FEA_IDX(FEA_DIVINED_CONFLICT_WEREWOLF);
                        if(r == Role::POSSESSED){
                            iterate(info.roleAlive(DAY_FIRST, Role::WEREWOLF), [&](int p)->void{ // 人狼全員で検討
                                if(info.divineResult[player][p] == Species::HUMAN){
                                    FooC(FEA_DIVINED_CONFLICT_WEREWOLF, base);
                                }
                            });
                        }else if(r == Role::SEER){
                            iterate(info.roleAlive(DAY_FIRST, Role::WEREWOLF), [&](int p)->void{ // 人狼全員で検討
                                if(info.divineResult[player][p] == Species::HUMAN){
                                    FooC(FEA_DIVINED_CONFLICT_WEREWOLF, base + 1);
                                }
                            });
                        }
                    }
                    
                    // 自分が人狼のときに、霊媒結果が違う
                    if(R == Role::WEREWOLF){
                        constexpr int base = FEA_IDX(FEA_INQUESTED_CONFLICT_WEREWOLF);
                        if(r == Role::POSSESSED){
                            iterate(info.roleAlive(DAY_FIRST, Role::WEREWOLF), [&](int p)->void{ // 人狼全員で検討
                                if(info.inquestResult[player][p] == Species::HUMAN){
                                    FooC(FEA_INQUESTED_CONFLICT_WEREWOLF, base);
                                }
                            });
                        }else if(r == Role::MEDIUM){
                            iterate(info.roleAlive(DAY_FIRST, Role::WEREWOLF), [&](int p)->void{ // 人狼全員で検討
                                if(info.inquestResult[player][p] == Species::HUMAN){
                                    FooC(FEA_INQUESTED_CONFLICT_WEREWOLF, base + 1);
                                }
                            });
                        }
                    }
                    
                    // 占い師が、自分の霊媒結果と矛盾することを報告した
                    if(R == Role::MEDIUM){
                        constexpr int base = FEA_IDX(FEA_DIVINED_CONFLICT_MY_INQUEST);
                        if(info.comingOutRole[player] == Role::SEER){
                            for(int p = 0; p < P; ++p){
                                if(info.divineResult[player][p] != NONE_SPECIES
                                   && info.divineResult[player][p] != getSpecies(info.playerRole(p))){
                                    FooC(FEA_DIVINED_CONFLICT_MY_INQUEST, base + r); break;
                                }
                            }
                        }
                    }
                    
                    // 霊媒師が、自分の占い結果と矛盾することを報告した
                    if(R == Role::SEER){
                        constexpr int base = FEA_IDX(FEA_INQUESTED_CONFLICT_MY_DIVINE);
                        if(info.comingOutRole[player] == Role::MEDIUM){
                            for(int p = 0; p < P; ++p){
                                if(info.inquestResult[player][p] != NONE_SPECIES
                                   && info.inquestResult[player][p] != getSpecies(info.playerRole(p))){
                                    FooC(FEA_INQUESTED_CONFLICT_MY_DIVINE, base + r); break;
                                }
                            }
                        }
                    }
                    
                    // 狼視点の襲撃失敗
                    /*if(R == Role::WEREWOLF){
                     constexpr int base = FEA_IDX(FEA_ATTACK_FAILURE_WEREWOLF);
                     if(r == Role::BODYGUARD){
                     for(int d = DAY_ATTACK_FIRST; d < day; ++d){
                     if(info.NRoleAlive(Role::WEREWOLF) <= 1){ break; } // 単独襲撃しかできない
                     if(!info.anyAttacked(d)){ // 襲撃が起こらなかった
                     if(!(info.attackVotedNum & 0xEEEEEEEEEEEEEEEE) // どの候補にも2票集まらなかった
                     && info.attackVotedNum.sum() > 1){ // 2人以上が投票した
                     if(info.isAliveAfterExecution(d, player)){
                     // 襲撃時点で生きていた場合に狩人であるか
                     // 単独襲撃失敗した場合は単純な計算なので学習しない
                     // 複数候補襲撃だった場合を考える
                     FooC(FEA_ATTACK_FAILURE, base);
                     }
                     }
                     }
                     }
                     }*/
                    
                    // 確定人狼に投票したかどうか
                    {
                        constexpr int base = FEA_IDX(FEA_VOTE_WEREWOLF);
                        
                    }
                }
                
            END_CANDIDATE:
                double exps = exp(s / est.temperature());
                FASSERT(exps, cerr << "s = " << s << " T = " << est.temperature() << endl;);
                est.template feedCandidateScore<M>(exps);
                
                dst[r] = exps;
            }
        }
        
#undef FooXC
#undef FooC
#undef FooX
#undef Foo
        
        template<int R, int M = 1, class info_t, class estimator_t>
        void calcEstimationScoreAll(double dst[N_MAX_PLAYERS][N_ROLES], const info_t& info, int myId, const estimator_t& est){
            for(int p = 0; p < info.rule().players; ++p){
                calcEstimationScore<R, M>(dst[p], info, myId, p, est);
            }
        }
        template<int M = 1, class info_t, class estimator_t>
        void calcEstimationScoreAllWithRole(double dst[N_MAX_PLAYERS][N_ROLES],
                                            const info_t& info, int myId, int myRole, const estimator_t& est){
            switch(myRole){
                case 0:{
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<0, M>(dst[p], info, myId, p, est);
                    }
                }break;
                case 1: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<1, M>(dst[p], info, myId, p, est);
                    }
                }break;
                case 2: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<2, M>(dst[p], info, myId, p, est);
                    }
                }break;
                case 3: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<3, M>(dst[p], info, myId, p, est);
                    }
                }break;
                case 4: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<4, M>(dst[p], info, myId, p, est);
                    }
                }break;
                case 5: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<5, M>(dst[p], info, myId, p, est);
                    }
                }break;
                default: {
                    for(int p = 0; p < info.rule().players; ++p){
                        calcEstimationScore<NONE_ROLE, M>(dst[p], info, myId, p, est);
                    }
                }break;
            }
            
        }
        
        template<class info_t, class sbjInfo_t, class estimator_t>
        void estimateNow(sbjInfo_t *const psi, int player, const info_t& info, const estimator_t& est){
            // 予測役職分布等の主観的な情報を設定
            calcEstimationScoreAllWithRole<0>(psi->estPD, info, player, info.playerRole(player), est);
            // 確実な主観情報の利用
            auto roleNum = info.rule().rolePlayers;
            BitArray64<4, 16> ret(0);
            auto seq = prepareWithCertainInfoWithRole(psi->estPD, ret, roleNum, info, player, info.playerRole(player));
            
            //cerr << "seq = " << seq << endl;
            
            // 役職決定していないプレーヤーを正規化
            normalize2(psi->estPD, 4, roleNum.sum(), seq, roleNum);
        }
        
        template<class info_t, class estimator_t>
        void estimateEachNow(info_t *const pinfo, int myId, const estimator_t& est){
            // 自分以外の全員、現局面の推定を行う
            for(int p = p; p < pinfo->rule().players; ++p){
                if(p != myId && pinfo->isAlive(pinfo->day(), p)){
                    pinfo->sub[p].estimateNow(p, est);
                }
            }
        }
    }
}

#endif // AIWOLF_THOUSAND_SINGLEESTIMATION_HPP_