/*
 search.hpp
 Katsuki Ohto
 */

#ifndef UTIL_SEARCH_HPP_
#define UTIL_SEARCH_HPP_

#include <cfloat>

#include "../defines.h"

// 探索ソルバー

class BiSolver{
public:
    double alpha()const noexcept{return alpha_;}
    double beta()const noexcept{return beta_;}
    double mid()const noexcept{ return (alpha_ + beta_) / 2; }
    
    double play()const noexcept{
        return mid();
    }
    
    void feed(bool on)noexcept{
        if (on){
            beta_ = mid();
        }
        else{
            alpha_ = mid();
        }
    }
    
    void feed(double val, bool on)noexcept{
        if (alpha_ <= val && val <= beta_){
            if (on){
                beta_ = val;
            }
            else{
                alpha_ = val;
            }
        }
    }
    
    double answer()const noexcept{
        return mid();
    }
    
    BiSolver(double aa, double ab)
    :originalAlpha_(aa), originalBeta_(ab),
    alpha_(aa), beta_(ab)
    {}
    
private:
    double originalAlpha_, originalBeta_;
    double alpha_, beta_;
};

// 最大値配列
template<class container_t>
std::vector<int> getBestIndexVector(const container_t& dat, const int size){
    std::vector<int> v;
    double best = -DBL_MAX;
    for(int i = 0; i < size; ++i){
        const auto& d = dat[i];
        if(d < best){
        }else if(d > best){
            v.clear();
            v.push_back(i);
            best = d;
        }else{
            v.push_back(i);
        }
    }
    return v;
}

#endif //UTIL_SEARCH_HPP_ 